"""
Watchdog timer.

This app works with a hardware watchdog. It falls back to software if necessary.

There is one main "wdt" option which

* hw

  Use/control the hardware watchdog. If set, ot

* timeout

  Main control. Cannot be turned off.

* external

  Require external keepalives. Reboot if they're missing.
"""

from __future__ import annotations

from moat.lib.micro import Event, L
from moat.lib.rpc import BaseCmd
from moat.micro.wdt import WDT, M

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


class Cmd(BaseCmd):
    """
    Watchdog Timer control.
    """

    wdt = None
    timeout = None

    doc = dict(
        _c=dict(
            _d="Watchdog control",
            hw="bool:use hardware watchdog=",
            ext="bool:require external keepalive",
            timeout="int:timeout (ms)",
            tt="int:tickle timer (ms, timeout/2)",
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        self._ping = Event()
        if cfg.get("hw", False) and M.WDT is not None:
            self.wdt = M.WDT
        else:
            self.wdt = WDT(cfg)

    async def task(self):  # noqa:D102
        self.wdt.setup()
        if L:
            self.set_ready()
        await self.wdt.run()

    async def reload(self):  # noqa:D102
        await super().reload()
        self.wdt.setup()
        self.wdt.ping()

    doc_x = dict(_d="feed watchdog", f="bool:force")

    async def cmd_x(self, f=False, n: Any = None):
        """
        External keepalive.

        @f: "force" param of `WDT.ping`.

        @n: ignored, used to distinguish calls during testing.
        """
        n  # noqa:B018
        self.wdt.ping(force=f)

    doc_set = dict(_d="set timeout", t="int:timeout,msec")

    async def cmd_set(self, t=None):
        """
        Set the watchdog timeout.

        Returns the previous value.
        """
        return self.wdt.tmo(t)

    doc_info = dict(_d="get state", _r=dict(t="int:timeout", x="bool:external?", h="bool:running"))

    async def cmd_info(self):
        """
        Current watchdog state::

            t: timeout
            x: external cfg
            h: hot?
        """
        if self.wdt is None:
            return None
        return dict(t=self.wdt.timeout, x=self.wdt.cfg["ext"], h=(self.wdt.wdt is not None))



# ruff:noqa:D100,F821,SLF001,I002

try:
    freeze("$(PORT_DIR)/modules")
except Exception:  # noqa:S110
    pass

require("copy")
require("collections")
require("contextlib")
require("functools")
require("inspect")
require("pprint")
require("types")
require("typing")
require("__future__")

include("$(MPY_DIR)/extmod/asyncio")

import moat.micro._embed.lib  # noqa:E402

# TODO only if building with networking
require("webrepl")

for p in moat.micro._embed.lib.__path__:
    print(p)
    include(p + "/manifest.py")



"""
Access a satellite's SPI bus (Linux/CPython version using spidev).
"""

from __future__ import annotations

from importlib import import_module

from moat.lib.micro import Lock, to_thread
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Protocol

    class _SpiDevProto(Protocol):
        max_speed_hz: int
        mode: int
        bits_per_word: int

        def open(self, bus: int, dev: int) -> None: ...

        def close(self) -> None: ...

        def readbytes(self, n: int) -> list[int]: ...

        def writebytes(self, data: list[int]) -> object: ...

        def xfer(self, data: list[int]) -> list[int]: ...

    class _SpiDevMod(Protocol):
        SpiDev: type[_SpiDevProto]


spidev = cast("_SpiDevMod", import_module("spidev"))

# Per-bus locks to prevent concurrent access
_bus_locks: dict[tuple[int, int], Lock] = {}


def _get_bus_lock(bus: int, device: int) -> Lock:
    """Get or create a lock for the given bus/device."""
    key = (bus, device)
    if key not in _bus_locks:
        _bus_locks[key] = Lock()
    return _bus_locks[key]


class Cmd(BaseCmd):
    r"""
    This command implements basic access to a SPI bus on Linux.

    Config::

        bus: 0      \# SPI bus number
        dev: 0      \# SPI device number (directly controls CS)
        f: 1000000  \# frequency, Hz (max_speed_hz)
        mode: 0     \# SPI mode (0-3, combines polarity and phase)
        bits: 8     \# bits per word

    On Linux, CS is controlled by the kernel driver based on the device number,
    so no separate CS path is needed.
    """

    _spi: _SpiDevProto

    doc = dict(
        _c=dict(_d="SPI driver (Linux)"),
        bus="int:bus number",
        dev="int:device number",
        f="int:frequency",
        mode="int:SPI mode(0-3)",
        bits="int:bits per word(8)",
    )

    def __init__(self, cfg: dict):
        super().__init__(cfg)

    async def setup(self):
        """
        Open a bus.
        """
        await super().setup()
        self._setup()

    async def reload(self):
        "reconfigured"
        self._teardown()
        self._setup()
        await super().reload()

    def _setup(self):
        cfg = self.cfg
        bus = cfg.get("bus", 0)
        dev = cfg.get("dev", 0)
        f = cfg.get("f", 1000000)
        mode = cfg.get("mode", 0)
        bits = cfg.get("bits", 8)

        self._spi = spi = cast("_SpiDevProto", spidev.SpiDev())
        spi.open(bus, dev)
        spi.max_speed_hz = f
        spi.mode = mode
        spi.bits_per_word = bits

        self._bus_id = (bus, dev)
        self.lock = _get_bus_lock(bus, dev)

    async def teardown(self):
        "shutdown"
        self._teardown()
        await super().teardown()

    def _teardown(self):
        s = self._spi
        if s is not None:
            del self._spi
            s.close()

    doc_rd = dict(
        _d="read",
        n="int:nbytes",
    )

    async def cmd_rd(self, n: int) -> bytes:
        """
        Read @n bytes from SPI.

        Args:
            n: number of bytes to read
        """
        async with self.lock:
            return bytes(await to_thread(self._spi.readbytes, n))

    doc_wr = dict(
        _d="write",
        buf="bytes:data",
        _r="int:nbytes written",
    )

    async def cmd_wr(self, buf: bytes) -> int:
        """
        Write @buf to SPI.

        Args:
            buf: data to write
        """
        async with self.lock:
            await to_thread(self._spi.writebytes, list(buf))
            return len(buf)

    doc_rw = dict(
        _d="write+read simultaneous",
        wbuf="bytes:write data",
        _r="bytes:read result",
    )

    async def cmd_rw(self, wbuf: bytes) -> bytes:
        """
        Simultaneous write and read (xfer).

        Args:
            wbuf: data to write

        Returns the same number of bytes as written.
        """
        async with self.lock:
            return bytes(await to_thread(self._spi.xfer, list(wbuf)))

    doc_wrrd = dict(
        _d="write then read (separate)",
        wbuf="bytes:write data",
        n="int:read bytes",
        _r="bytes:read result",
    )

    async def cmd_wrrd(self, wbuf: bytes, n: int) -> bytes:
        """
        Write @wbuf then read @n bytes (sequential, not simultaneous).

        This is useful for command-response protocols where you send
        a command and then read the response.

        Args:
            wbuf: data to write first
            n: number of bytes to read after
        """
        spi = self._spi
        async with self.lock:
            await to_thread(spi.writebytes, list(wbuf))
            return bytes(await to_thread(spi.readbytes, n))



"""
Access a satellite's file system.
"""

from __future__ import annotations

import errno
import os

from moat.lib.codec.errors import FileExistsError, FileNotFoundError  # noqa:A004
from moat.lib.micro import to_thread
from moat.lib.rpc import LockBaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.util import attrdict

    from typing import Protocol

    class _FileObj(Protocol):
        def close(self) -> object: ...

        def seek(self, pos: int) -> object: ...

        def read(self, n: int) -> bytes: ...

        def write(self, data: bytes) -> int: ...


def _fty(s, **r):
    # file type/size: stat array to dict
    sn = s[0]
    if sn & 0x8000:  # file
        m = "f"
    elif sn & 0x4000:  # file
        m = "d"
    else:
        m = "?"
    r["m"] = m

    if m == "f":
        r["s"] = s[6]
    if m != "?":
        r["t"] = s[7]
    return r


def _efix(f, p, *a):
    # fix errors
    try:
        return f(p, *a)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise FileNotFoundError(p) from None
        if e.errno == errno.EEXIST:
            raise FileExistsError(p) from None
        raise


class Cmd(LockBaseCmd):
    """
    File system access.

    Set "root" to the file system path this command should apply to.
    """

    _fd_last = 0
    _fd_cache: dict[int, _FileObj]

    doc = dict(_c=dict(_d="Local file system access", root="str:root path"))

    def __init__(self, cfg: attrdict):
        super().__init__(cfg)
        self._fd_cache = dict()
        try:
            self._pre = cfg["root"]
        except KeyError:
            self._pre = ""
        else:
            if self._pre and self._pre[-1] != "/":
                self._pre += "/"

    def _fsp(self, p: str):
        if self._pre:
            p = self._pre + p
        if p == "":
            p = "/"
        # elif p == ".." or p.startswith("../") or "/../" in p: or p.endswith("/..")
        # raise FSError("nf")
        return p

    def _fd(self, fd: int):
        # filenr > fileobj
        f = self._fd_cache[fd]
        # log(f"{fd}:{repr(f)}")
        return f

    def _add_f(self, f):
        # fileobj > filenr
        self._fd_last += 1
        fd = self._fd_last
        self._fd_cache[fd] = f
        # log(f"{fd}={repr(f)}")
        return fd

    async def _del_f(self, fd: int):
        f = self._fd_cache.pop(fd)
        # log(f"{fd}!{repr(f)}")
        await to_thread(f.close)

    doc_reset = dict(_d="close all", p="str:set new prefix")

    async def cmd_reset(self, p: str | None = None):
        "close all"
        for v in self._fd_cache.values():
            await to_thread(v.close)
        self._fd_cache = dict()

        if not p or p == "/":
            self._fs_prefix = ""
        elif p[0] == "/":
            self._fs_prefix = p
        else:
            self._fs_prefix += "/" + p

    doc_open = dict(_d="open file", _0="str:path", m="str:mode (r,w)", _r="int:fileid")

    async def cmd_open(self, p: str, m: str = "r"):
        "open @f in binary mode @m (r,w)"
        p = self._fsp(p)
        f = await to_thread(_efix, open, p, m + "b")
        return self._add_f(f)

    doc_rd = dict(_d="read file", _0="int:fileid", _1="int:offset", n="int:length")

    async def cmd_rd(self, f: int, o: int = 0, n: int = 64):
        "read @n bytes from @f at offset @o"
        fh = self._fd(f)
        fh.seek(o)
        return await to_thread(fh.read, n)

    doc_wr = dict(_d="write file", _0="int:fileid", _1="int:offset", d="bytes:data")

    async def cmd_wr(self, f: int, o: int = 0, d: bytes = b"") -> int:
        "write @d to @f at offset @o"
        fh = self._fd(f)
        fh.seek(o)
        return await to_thread(fh.write, d)

    doc_cl = dict(_d="close file", _0="int:fileid")

    async def cmd_cl(self, f):
        "close @f"
        await self._del_f(f)

    doc_ls = dict(_d="list dir", _0="str:path", x="bool:return mapping", _o="str|dict")

    async def cmd_ls(self, p: str = "", x: bool = False):
        """
        dir of @p.

        Set @x to return a list of stat mappings::
        n: name
        t: time
        s: size
        m: type
        """
        p = self._fsp(p)
        res = []
        ilistdir = getattr(os, "ilistdir", None)
        if ilistdir is None:
            it = ((n,) for n in os.listdir(p))
        else:
            it = await to_thread(ilistdir, p)

        def inext(it):
            try:
                return next(it)
            except StopIteration:
                return (None,)

        if x:
            while True:
                n, *_ = await to_thread(inext, it)
                if n is None:
                    break
                st = await to_thread(os.stat, f"{p}/{n}")
                res.append(_fty(st, n=n))
        else:
            while True:
                n, *_ = await to_thread(inext, it)
                if n is None:
                    break
                res.append(n)
        return res

    doc_mkdir = dict(_d="make dir", _0="str:path")

    async def cmd_mkdir(self, p: str):
        "new dir at @p"
        p = self._fsp(p)
        await to_thread(_efix, os.mkdir, p)

    doc_hash = dict(_d="hash file", _0="str:path", l="int:prefixlen", _r="bytes:hash")

    async def cmd_hash(self, p: str, l: int | None = None):  # noqa: E741
        """
        Hash the contents of @p, sha256
        """
        import hashlib  # noqa: PLC0415

        p = self._fsp(p)

        def hash_file(p):
            h = hashlib.sha256()
            mem = memoryview(bytearray(512))
            with open(p, "rb") as f:
                while True:
                    n = f.readinto(mem)
                    if not n:
                        break
                    h.update(mem[:n])
            return h.digest()

        res = await to_thread(hash_file, p)
        if l is not None:
            res = res[:l]
        return res

    doc_stat = dict(
        _d="stat file",
        _0="str:path",
        v="bool:include struct",
        _r=dict(m="int:mode", s="int:size", t="int:modtime", d="list:state struct"),
    )

    async def cmd_stat(self, p: str, v: bool = False):
        """
        State of @p.

        Returns a mapping::

            m: mode (f,d,?)
            s: size (files only)
            t: mod time
            d: state array

        @d will not be sent if @v is False.
        """
        p = self._fsp(p)
        s = await to_thread(_efix, os.stat, p)
        res = _fty(s)
        if v:
            res["d"] = s
        return res

    doc_mv = dict(
        _d="rename/move file",
        _0="str:source",
        _1="str:dest",
        x="str:exchange temp",
        n="bool:dest must be new",
    )

    async def cmd_mv(self, s: str, d: str, x: str | None = None, n: bool = False):
        """
        move file @s to @d.

        If @n is True, the destination must not exist.

        @x is the name of a temp file; if set, it is used for swapping @s
        and @d.
        """
        p = self._fsp(s)
        q = self._fsp(d)
        os.stat(p)  # must exist
        if n:
            # dest must not exist
            try:
                _efix(os.stat, q)
            except FileNotFoundError:
                pass
            else:
                raise FileExistsError(q)
        if x is None:
            await to_thread(os.rename, p, q)
        else:
            x = self._fsp(x)
            # exchange contents, via third file
            try:
                _efix(os.stat, x)
            except FileNotFoundError:
                pass
            else:
                raise FileExistsError(q)
            await to_thread(os.rename, p, x)
            await to_thread(os.rename, q, p)
            await to_thread(os.rename, x, q)

    doc_rm = dict(_d="remove file", _0="str:path")

    async def cmd_rm(self, p: str):
        "unlink file @p"
        p = self._fsp(p)
        await to_thread(_efix, os.remove, p)

    doc_rmdir = dict(_d="remove dir", _0="str:path")

    async def cmd_rmdir(self, p: str):
        "unlink dir @p"
        p = self._fsp(p)
        await to_thread(_efix, os.rmdir, p)

    doc_new = dict(_d="create file", _0="str:path")

    async def cmd_new(self, p: str):
        "new file @p"
        p = self._fsp(p)
        f = await to_thread(_efix, open, p, "wb")
        await to_thread(f.close)



"""
This is a vendorized minimal version of the "outcome" package,
from https://pypi.org/project/outcome/.

Contrary to the upstream version, regular values (but not exceptions)
may be unwrapped more than once.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from typing import Any

__all__ = ["Error", "Outcome", "Value", "capture"]

T = TypeVar("T")


def capture(sync_fn: Callable[..., T], *args: Any, **kwargs: Any) -> Value[T] | Error:
    """Run ``sync_fn(*args, **kwargs)`` and capture the result."""
    try:
        return Value(sync_fn(*args, **kwargs))
    except Exception as exc:
        return Error(exc)


async def acapture(
    async_fn: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> Value[T] | Error:
    """Run ``await async_fn(*args, **kwargs)`` and capture the result."""
    try:
        return Value(await async_fn(*args, **kwargs))
    except Exception as exc:
        return Error(exc)


class Outcome:
    """An abstract class representing the result of a Python computation."""


class Value(Outcome, Generic[T]):
    """Concrete :class:`Outcome` subclass representing a regular value."""

    def __init__(self, value: T) -> None:
        self.value: T = value

    def __repr__(self) -> str:
        return f"Value({self.value!r})"

    def unwrap(self) -> T:  # noqa: D102
        return self.value


class Error(Outcome):
    """Concrete :class:`Outcome` subclass representing a raised exception."""

    def __init__(self, error: Exception) -> None:
        self.error: Exception = error

    def __repr__(self) -> str:
        return f"Error({self.error!r})"

    def unwrap(self) -> None:  # noqa: D102
        captured_error = self.error
        try:
            raise captured_error
        finally:
            del captured_error, self



# MIT license; Copyright (c) 2022 Damien P. George  # noqa: D100
# © 2025 Matthias Urlichs
from __future__ import annotations

import anyio
import os

try:
    from mpy_cross import run as mpy_cross_run
except ImportError:
    mpy_cross_run = None


class VfsRomWriter:  # noqa: D101
    ROMFS_HEADER = b"\xd2\xcd\x31"

    ROMFS_RECORD_KIND_UNUSED = 0
    ROMFS_RECORD_KIND_PADDING = 1
    ROMFS_RECORD_KIND_DATA_VERBATIM = 2
    ROMFS_RECORD_KIND_DATA_POINTER = 3
    ROMFS_RECORD_KIND_DIRECTORY = 4
    ROMFS_RECORD_KIND_FILE = 5

    def __init__(self, cross):
        self._dir_stack = [(None, bytearray())]
        self.cross = cross

    def _encode_uint(self, value):
        encoded = [value & 0x7F]
        value >>= 7
        while value != 0:
            encoded.insert(0, 0x80 | (value & 0x7F))
            value >>= 7
        return bytes(encoded)

    def _pack(self, kind, payload):
        return self._encode_uint(kind) + self._encode_uint(len(payload)) + payload

    def _extend(self, data):
        buf = self._dir_stack[-1][1]
        buf.extend(data)
        return len(buf)

    def finalise(self):  # noqa: D102
        _, data = self._dir_stack.pop()
        encoded_kind = VfsRomWriter.ROMFS_HEADER
        encoded_len = self._encode_uint(len(data))
        if (len(encoded_kind) + len(encoded_len) + len(data)) & 1:
            encoded_len = b"\x80" + encoded_len
        data = encoded_kind + encoded_len + data
        return data

    def opendir(self, dirname):  # noqa: D102
        self._dir_stack.append((dirname, bytearray()))

    def closedir(self):  # noqa: D102
        dirname, dirdata = self._dir_stack.pop()
        dirdata = self._encode_uint(len(dirname)) + bytes(dirname, "utf-8") + dirdata
        self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DIRECTORY, dirdata))

    def mkdata(self, data):  # noqa: D102
        assert len(self._dir_stack) == 1
        return self._extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, data)) - len(
            data
        )

    def mkfile(self, filename, filedata):  # noqa: D102
        filename = bytes(filename, "utf-8")
        payload = self._encode_uint(len(filename))
        payload += filename
        if isinstance(filedata, tuple):
            sub_payload = self._encode_uint(filedata[0])
            sub_payload += self._encode_uint(filedata[1])
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_POINTER, sub_payload)
        else:
            payload += self._pack(VfsRomWriter.ROMFS_RECORD_KIND_DATA_VERBATIM, filedata)
        self._dir_stack[-1][1].extend(self._pack(VfsRomWriter.ROMFS_RECORD_KIND_FILE, payload))

    async def copy_in():  # noqa: D102
        pass

    async def copy_recursively(self, src_dir):  # noqa: D102
        DIR = 1 << 14
        dir_contents = sorted(os.listdir(src_dir))
        for name in dir_contents:
            src_name = src_dir / name
            st = os.stat(src_name)

            if st[0] & DIR:
                # A directory, enter it and copy its contents recursively.
                self.opendir(name)
                await self.copy_recursively(src_name)
                self.closedir()
            else:
                # A file.
                did_mpy = False
                if self.cross and name.endswith(".py"):
                    name_mpy = name[:-3] + ".mpy"
                    src_name_mpy = src_dir / name_mpy
                    if not await anyio.Path(src_name_mpy).is_file():
                        if mpy_cross_run is not None:
                            did_mpy = True
                            mpy_cross_run(src_name)
                if did_mpy:
                    name = name_mpy  # noqa:PLW2901
                    src_name = src_name_mpy
                src = await anyio.Path(src_name).read_bytes()
                self.mkfile(name, src)


async def make_romfs(src: anyio.Path, mpy_cross):  # noqa: D103
    vfs = VfsRomWriter()

    # Build the filesystem recursively.
    await vfs.copy_recursively(src, mpy_cross)
    return vfs.finalise()


async def write_romfs(dst, src: anyio.File):  # noqa: D103
    raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Self

    ByteType = bytes | bytearray | memoryview
    VarByteType = bytearray | memoryview


class NoCodecError(ValueError):
    "No codec found"


class IncompleteData(ValueError):
    "Stream chopped off"


class Codec:
    """
    This is an abstract base class. It supports block and
    incremental decoders.

    To support block decoding, override the decode method.

    Incremental codecs need to support interleaved operation. This means
    that the decoder must support two modes.

    The initial mode is to append all data passed in via `feed` to an
    internal buffer. If necessary, the application then uses `unfeed` to
    read data off the start of the buffer until a packet header is
    detected.

    After the header is consumed, the application must repeat calling
    ``__anext__`` until an object is returned. (If the decoder
    raises `StopAsyncIteration`, feed more data to it and repeat.)

    The decoder must leave all incoming data extending past the object in
    its buffer. It must not attempt to decode them until the next call to
    ``__anext__``.
    """

    def __init__(self, ext=None):
        if ext is None:
            ext = Extension()  # empty
        self.ext = ext
        self.buf = b""

    def copy(self) -> Self:
        """
        Returns a new codec with identical configuration.
        """
        raise NotImplementedError

    def encode(self, obj: Any) -> ByteType:
        """
        Encode a data structure, yielding some bytes which may or may not
        be self-terminating.
        """
        raise NotImplementedError

    def decode(self, data: ByteType) -> Any:
        """
        Decode a block of data, which must result in a single message.
        """
        raise NotImplementedError

    def feed(self, data: ByteType) -> None:
        """
        Add to the codec's buffer.

        This method buffers input data (unless they belong to an
        object-in-progress). If there is no such object, this method must
        add all data to the buffer. The same holds for data that don't
        belong to the current object.
        """
        raise NotImplementedError

    def __iter__(self):
        return self

    def __next__(self):
        """
        Decode the next item in the input stream and return it.

        After calling this method, the decoder has an object in progress if
        (only if) `StopIteration` was raised.
        """
        raise NotImplementedError

    def unfeed(self, buf: VarByteType) -> int:
        """
        Take from the front of the decoder's buffer.

        This method may not be called while there is an object in progress.
        """
        raise NotImplementedError


if TYPE_CHECKING:
    EncoderFunc = Callable[[Codec, Any], ByteType | tuple[int, ByteType]]
    DecoderFunc = Callable[[Codec, ByteType], Any]


class Extension:
    def __init__(self):
        self.enc: dict[type, tuple[int | None, Callable]] = {}
        self.dec: dict[int, Callable] = {}

    def copy(self):
        res = type(self)()
        res.enc.update(self.enc)
        res.dec.update(self.dec)
        return res

    @overload
    def encoder(self, key: int | None, cls: type) -> Callable[[EncoderFunc], EncoderFunc]: ...

    @overload
    def encoder(self, key: int | None, cls: type, fn: EncoderFunc) -> EncoderFunc: ...

    def encoder(
        self, key: int | None, cls: type, fn: EncoderFunc | None = None
    ) -> EncoderFunc | Callable[[EncoderFunc], EncoderFunc]:
        def _enc(fn: EncoderFunc) -> EncoderFunc:
            self.enc[cls] = (key, fn)
            return fn

        if fn is None:
            return _enc
        return _enc(fn)

    @overload
    def decoder(self, key: int | None) -> Callable[[DecoderFunc], DecoderFunc]: ...

    @overload
    def decoder(self, key: int | None, fn: DecoderFunc) -> DecoderFunc: ...

    def decoder(
        self, key: int | None, fn: DecoderFunc | None = None
    ) -> DecoderFunc | Callable[[DecoderFunc], DecoderFunc]:
        def _dec(fn: DecoderFunc) -> DecoderFunc:
            self.dec[key] = fn  # type: ignore[index]  # key can be None for catch-all
            return fn

        if fn is None:
            return _dec
        return _dec(fn)

    def encode(self, codec: Codec, obj) -> tuple[int, ByteType]:
        try:
            key, fn = self.enc[type(obj)]
        except KeyError:
            try:
                key, fn = self.enc[object]
            except KeyError:
                raise NoCodecError(codec, obj) from None

        res = fn(codec, obj)
        if key is None:
            key, res = res
        return key, res

    def decode(self, codec: Codec, key: int, data: ByteType) -> Any:
        try:
            fn = self.dec[key]
        except KeyError:
            raise NoCodecError(codec, key) from None
        return fn(codec, data)



"""
Access any moat.micro Buf/Blk/Msg device
"""

from __future__ import annotations

from moat.util import import_
from moat.lib.micro import AC_use
from moat.lib.rpc import BaseCmdBBM


# Serial packet forwarder
# cfg:
# uart: N
# tx: PIN
# rx: PIN
# baud: 9600
# max:
#   len: N
#   idle: MSEC
# start: NUM
#
class Port(BaseCmdBBM):
    """
    Access any moat.micro Buf/Blk/Msg device.

    The config item "device" must contain the class path.
    """

    doc = dict(_c=dict(_d="device access", device="str:class path", _k="any:class setup"))

    pack = None

    async def stream(self):
        "setup,"
        intf = import_(self.cfg["device"])
        return await AC_use(self, intf(self.cfg))



"""
Raw serial port app.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.stream.cmdbbm import BaseCmdBBM
from moat.micro.app._doc import _mode_d

from ._util import get_serial


class Raw(BaseCmdBBM):
    """Sends/receives raw bytes off a serial port."""

    doc = dict(_c=dict(_d="raw serial data", port="str:Port or 'USB'", mode=_mode_d))
    pack = None

    async def stream(self):
        """Returns the serial stream."""
        return await AC_use(self, get_serial(self.cfg))



"""
Stream support for MoaT RPC.
"""

from __future__ import annotations

# Re-export everything from base for backward compatibility
from .base import *  # noqa: F403

# Import and re-export from cmdbbm, xcmd
from .cmdbbm import *  # noqa: F403
from .xcmd import *  # noqa: F403

__all__ = []

# Collect __all__ from all submodules
from . import base, cmdbbm, xcmd

for mod in (base, cmdbbm, xcmd):
    if hasattr(mod, "__all__"):
        __all__ += mod.__all__



"""
Access a satellite's i²c bus.
"""

from __future__ import annotations

from functools import partial

try:
    import machine
except ImportError:
    from moat.micro._test import machine

import contextlib

from moat.lib.micro import Lock, to_thread
from moat.lib.rpc import BaseCmd


class Cmd(BaseCmd):
    """
    This command implements basic access to an I²C bus.

    Config::

        id: None  # use soft i2c
        c: 3  # pin# of control
        d: 4  # pin# of data
        cx: {}  # additional machine.Pin params for control pin
        dx: {}  # additional machine.Pin params for data pin
        f: 100000  # frequency, Hz
        t: 1000 # bus timeout, msec
    """

    _bus = None

    doc = dict(
        _c=dict(_d="i²c driver"),
        id="int|None:hardware",
        c="int:ctrl pin",
        d="int:data pin",
        f="int:frequency",
        t="int:timeout(ms)",
        cx="dict:ctrl hw init",
        dx="dict:dta hw init",
    )

    async def setup(self):
        """
        Open a bus.
        """
        await super().setup()
        self._setup()

    async def reload(self):
        "reconfigured"
        self._teardown()
        self._setup()
        await super().reload()

    def _setup(self):
        cfg = self.cfg
        c = machine.Pin(cfg["c"], **cfg.get("cx", {}))
        d = machine.Pin(cfg["d"], **cfg.get("dx", {}))
        f = cfg.get("f", 100000)
        t = cfg.get("t", 1000) * 1000
        if (i := cfg.get("id", None)) is None:
            cls = machine.SoftI2C
        else:
            cls = partial(machine.I2C, i)
        self._bus = cls(scl=c, sda=d, freq=f, timeout=t)
        self.lock = Lock()

    async def teardown(self):
        "shutdown"
        self._teardown()
        await super().teardown()

    def _teardown(self):
        b, self._bus = self._bus, None
        if b is not None:
            with contextlib.suppress(AttributeError):
                b.deinit()

    doc_rd = dict(_d="read", _0="int:addr", n="int:nbytes(16)")

    async def cmd_rd(self, i, n=16) -> bytes:
        "read @n bytes from bus @cd at address @i"
        async with self.lock:
            return await to_thread(self._bus.readfrom, i, n)

    doc_wr = dict(_d="write", _0="int:addr", buf="bytes:data", _r="int:nbytes")

    async def cmd_wr(self, i, buf):
        "write @buf to bus @cd at address @i"
        async with self.lock:
            return await to_thread(self._bus.writeto, i, buf)

    doc_wrrd = dict(
        _d="write+read",
        _0="int:addr",
        buf="bytes:data",
        n="int:nbytes(16)",
        _r="int|bytes:nbytes short-written|read result",
    )

    async def cmd_wrrd(self, i, buf, n=16) -> bytes:
        """
        write @buf to bus @cd at address @i, then read @n bytes.

        Returns -x if only x bytes could be written.
        """

        def _run() -> bytes:
            bus = self._bus
            d = self._bus.writeto(i, buf, False)
            if d < len(buf):
                bus.stop()
                return d
            return self._bus.readfrom(i, n)

        async with self.lock:
            return await to_thread(_run)

    doc_scan = dict(_d="bus scan")

    async def cmd_scan(self) -> list[int]:
        """
        Scan the bus.

        Returns: list of valid addresses.
        """
        async with self.lock:
            return await to_thread(self._bus.scan)



"""
Broadcasting support
"""

from __future__ import annotations

from ._impl import Broadcaster as Broadcaster
from ._impl import BroadcastReader as BroadcastReader
from ._impl import LostData as LostData

__all__ = [
    "BroadcastReader",
    "Broadcaster",
    "LostData",
]



"""
Error handler and retry app.
"""

from __future__ import annotations

from moat.lib.rpc import RetryCmd


class Err(RetryCmd):
    """An error handler and possibly-retrying subcommand manager."""

    pass



"""
CBOR message encoding/decoding for stream layers.
"""

from __future__ import annotations

from moat.lib.micro import AC_use, Lock, log
from moat.lib.stream import BaseBuf, StackedMsg

from ._console import _CReader

from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from moat.lib.codec import Codec
    from moat.lib.stream.base import Buffer, MutBuffer
    from moat.util.liner import Liner

    from typing import Any


class _CBORMsgBuf(StackedMsg):
    """
    structured messages > CBOR bytestream

    Use this if your stream is reliable (TCP, USB, …) but doesn't support
    message boundaries.

    If @console is set and a prefix is used, sends data atomically.
    Otherwise two separate write calls are used to save on message copying.

    Config:
        console:
            Flag how to handle non-framed data.
            True: collect for crd/cwr, False: print incoming, None: ignore.
            If an integer: Buffer size
        msg_prefix:
            value of the prefix byte for messages (as opposed to console data).

    :meta public:
    """

    cons: bool | int = False
    codec: Codec
    liner: Liner | None = None

    def __init__(self, stream: BaseBuf, cfg: dict):
        StackedMsg.__init__(self, stream, cfg)
        self.w_lock = Lock()

        pref = cfg.get("msg_prefix")
        if pref is not None:
            pref = bytes((pref,))
        self.pref = pref

        cons = cfg.get("console", False)
        if cons:
            _CReader.__init__(cast(_CReader, self), cons)

    async def setup(self):
        await super().setup()
        if self.cons is False:
            from moat.util.liner import Liner  # noqa:PLC0415

            self.liner = await AC_use(self, Liner())

    async def cwr(self, buf: Buffer) -> None:
        if not self.cons:
            return
        await self.s.wr(buf)

    async def crd(self, buf: MutBuffer) -> int:
        return await _CReader.crd(cast(_CReader, self), buf)

    async def send(self, m: Any) -> Any:
        try:
            msg = self.codec.encode(m)
        except Exception:
            log("MSG:\n%r", m)
            raise
        async with self.w_lock:
            if self.pref is not None:
                if True:  # self.cons:
                    msg = self.pref + msg  # must be atomic
                else:
                    await self.s.wr(self.pref)
            await self.s.wr(msg)

    async def recv(self) -> Any:
        """
        Receive the next object.
        """
        # Pre+postcondition: the codec does not have an object in progress.

        codec = self.codec
        buf = bytearray(64)
        if self.pref is None:
            # easy case
            while True:
                try:
                    r = next(codec)
                except StopIteration:
                    n = await self.s.rd(buf)
                    codec.feed(memoryview(buf)[:n])
                else:
                    if self.cons and isinstance(r, int) and r >= 0:
                        _CReader.cput(cast(_CReader, self), r)
                    else:
                        return r

        while True:
            b = bytearray(1)
            # read until we get a prefix byte
            if codec.unfeed(b) == 0:
                n = await self.s.rd(buf)
                codec.feed(memoryview(buf)[:n])
            elif b == self.pref:
                break
            elif self.cons:
                _CReader.cput(cast(_CReader, self), b[0])
            elif self.liner is not None:
                self.liner(bytes(b))

        while True:
            # read until we get an object
            try:
                return next(codec)
            except StopIteration:
                pass

            n = await self.s.rd(buf)
            codec.feed(memoryview(buf)[:n])


class _CBORMsgBlk(StackedMsg):
    """
    structured messages > chunked bytestrings

    Use this if the layer below supports byte boundaries
    (one bytestring-ized message per call).

    :meta public:
    """

    codec: Codec

    async def send(self, m: Any) -> Any:
        await self.s.snd(self.codec.encode(m))

    async def recv(self):
        m = await self.s.rcv()
        return self.codec.decode(m)



"""
Raw remote stream forwarding app.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.stream.cmdbbm import BaseCmdBBM

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from types import CoroutineType

    from moat.lib.stream import BaseBuf

    from typing import Any


class Raw(BaseCmdBBM):
    """
    Link to a stream that's someplace else.

    This app forwards read/write requests to somewhere else.
    """

    doc = dict(_c=dict(_d="Data forwarding", path="path:dest"))

    def stream(self) -> CoroutineType[Any, Any, BaseBuf]:
        """Returns the link."""
        return AC_use(self, self.root.sub_at(self.cfg["path"]))



"""
fake sensors
"""

from __future__ import annotations

import random
from math import atanh, tanh

from moat.util import NotGiven
from moat.lib.micro import Event, TaskGroup
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import EllipsisType

PINS = {}


class Pin(BaseCmd):
    """
    This is a fake Pin.

    Iterating it yields a new value whenever the pin changes.
    """

    doc = dict(_c=dict(_d="A fake I/O pin", pin="int:pin#"))

    flag: Event

    def __init__(self, cfg):
        super().__init__(cfg)
        PINS[cfg["pin"]] = self
        self._value = cfg.get("init", False)
        self.flag = Event()

    def in_value(self, val):
        "set+send pin value unconditionally"
        self.flag.set()
        self.flag = Event()
        self._value = val

    @property
    def value(self):
        "current pin value"
        return self._value

    def __aiter__(self):
        return self

    async def get(self):
        "Wait for + get the next value"
        await self.flag.wait()
        return self._value

    async def cmd(self, val: bool | EllipsisType = NotGiven) -> None | bool:
        "Simple Data protocol."
        if val is NotGiven:
            return self.value
        self.in_value(val)

    async def stream(self, msg):
        """
        R/W data stream. The initial argument says whether you want to
        read. Waits for change if @o (old value) is not None.
        """
        o = msg.get("o", None)
        async with TaskGroup() as tg, msg.stream() as m:

            async def _rd():
                val = self.value
                if o is not val:
                    await m.send(val)
                while True:
                    val = await self.get()
                    await m.send(val)

            if o is not None or msg.get(0, False):
                tg.start_soon(_rd)

            for mm in m:
                self.in_value(mm[0])
            tg.cancel()
            await msg.result()


class ADC(BaseCmd):
    """
    This is a "fake" ADC that walks between a given min and max value.

    The min/max boundary values will never be returned.

    Config parameters:
    - min, max: range. Defaults to 0…1.
    - step: max difference between two consecutive values.
    - border: A hint for how long the sequence should be close to
      the min/max. Float. Default 2.
    - seed: used to reproduce the random sequence.
    """

    doc = dict(
        _c=dict(
            _d="A random analog input",
            min="float:min value",
            max="float:max value",
            border="float:border",
            seed="int:random seed",
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        cfg = self.cfg
        self.min = cfg.get("min", 0)
        self.max = cfg.get("max", 1)
        self.border = cfg.get("border", 2)
        self.step = cfg["step"] / (self.max - self.min) / 2 if "step" in cfg else 0.1

        self.val = (
            atanh(((cfg.init - self.min) / (self.max - self.min) - 0.5) * 2)
            if "init" in cfg
            else 0
        )
        self.bias = 0
        self.rand = random.Random(cfg.get("seed", None))

    doc_r = dict(_d="read")

    async def cmd_r(self):
        "read current value"
        b = self.bias + (self.rand.random() - 0.5) * self.step
        v = self.val + b
        if v > self.border and b > 0:
            b = 0
        elif v < -self.border and b < 0:
            b = 0

        self.val = v
        self.bias = b

        return self.min + (self.max - self.min) * (0.5 + 0.5 * tanh(v))



"""
# MsgHandler on top of a message stream

MoaT-Cmd offers no way to discover who your caller is. This is sometimes
inconvenient.

As a real-world example, assume you have an embedded device that's
connected to some environmental sensors and a ventilator, so it needs to
calculate the air quality. Assume further that the library to do this
is too large, or closed-source with binaries not available for its
architecture (hello, Bosch, please reconsider), or single-theaded but the
embedded device doesn't have threads.

So you use MoaT-Cmd to call out to your server, which has the library. The
library then wants to access your embedded device's i²c bus to actually
read the data, but it doesn't have its address.

The solution is to multiplex the stream to the server … with error handling,
streaming the resulting measurements, and all that. But why re-invent the
wheel, when MoaT-Cmd already *is* a multiplexing library, and you're using
it anyway?

Hence this stream handler.

See ``tests/moat_lib_rpc/test_nest.py`` for an example.

"""

from __future__ import annotations

from moat.lib.micro import ACM, AC_exit, log

from .stream import HandlerStream

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from moat.lib.rpc import Msg

    from .base import BaseMsgHandler

__all__ = ["CmdStream", "rpc_on_rpc"]


class CmdStream(HandlerStream):
    """
    This command stream uses the data from a single message as its
    transport.

    Args:
        cmd: the command handler to call for incoming commands.
             May be ``None`` if you don't handle any.
        msg: the stream to use. **It must be wrapped in an `async with
             msg.stream()` block.**
        debug: Prefix for tracing. Note that the trace handles raw
               message data and does not decode transactions.
    """

    __msg: Msg

    def __init__(
        self,
        cmd: BaseMsgHandler | None,
        msg: Msg,
        debug: str | None = None,
        **kw,
    ):
        self.__msg = msg
        self.__debug = debug

        super().__init__(cmd, **kw)

    async def read_stream(self):  # noqa: D102
        msg = self.__msg

        async for m in msg:
            if m.kw:
                log("R%s: incoming keywords ignored!? %r", self.__debug or "", m)
            elif self.__debug:
                log("R%s %r", self.__debug, m)
            await self.msg_in(cast("list", m.args_l))

    async def write_stream(self):  # noqa: D102
        msg = self.__msg
        while True:
            try:
                m = await self.msg_out()
            except EOFError:
                return
            if self.__debug:
                log("W%s %r", self.__debug, m)

            await msg.send(*m)


class rpc_on_rpc:
    """
    Run a command handler on top of a message stream @msg.
    """

    def __init__(
        self,
        cmd: BaseMsgHandler,
        msg: Msg,
        *,
        debug: bool = False,
        logger=None,
    ):
        """
        Args:
            cmd: handler for incoming messages. May be `None`.
            msg: MoaT-RPC Transport stream
            debug: flag whether to emit a message debug trace
            logger: callable for debugging internal state

        This is an async context manager that yields a `CmdStream`.
        """
        self.cmd = cmd
        self.msg = msg
        self.debug = debug
        self.logger = logger

    async def __aenter__(self) -> CmdStream:
        AC = ACM(self)
        try:
            dbg = "" if self.debug else None
            return await AC(CmdStream(self.cmd, self.msg, debug=dbg, logger=self.logger))

        except BaseException as exc:
            await AC_exit(self, type(exc), exc, None)
            raise

    async def __aexit__(self, *exc):
        return await AC_exit(self, *exc)



"""
App to open a channel to a process.
"""

from __future__ import annotations

from moat.lib.micro import AC_use
from moat.lib.rpc.cmd.msg import BaseCmdMsg
from moat.lib.rpc.stream.cmdbbm import BaseCmdBBM
from moat.lib.stream import ProcessBuf, serial_stack


class ProcessCmd(BaseCmdMsg):
    """
    Channel to a process that handles MoaT messages
    """

    argv = None
    path = None

    doc = dict(
        _c=dict(
            _d="Command link to a subprocess",
            command="list[str]:argv",
            path="str:file of executable",
        )
    )

    async def stream(self):  # noqa:D102
        argv = self.cfg["command"] if self.argv is None else self.argv
        path = self.cfg.get("path") if self.path is None else self.path
        if path is None and argv[0][0] == "/":
            path = argv[0]

        proc = ProcessBuf(argv, executable=path)
        return await AC_use(self, serial_stack(proc, cfg=self.cfg))


class ProcessIO(BaseCmdBBM):
    """
    Byte channel to a process that handles arbitrary MaoT messages
    """

    argv = None
    path = None

    doc = dict(
        _c=dict(
            _d="Data link to a subprocess", command="list[str]:argv", path="str:file of executable"
        )
    )

    async def stream(self):  # noqa:D102
        argv = self.cfg["command"] if self.argv is None else self.argv
        path = self.cfg.get("path") if self.path is None else self.path
        if path is None and argv[0][0] == "/":
            path = argv[0]

        proc = ProcessBuf(argv, executable=path)
        return await AC_use(self, proc)



"""
"Slow" PWM handler.

This is intended for slow devices like a thermically controlled heating
valve, where switching every five seconds or so works perfectly well.
"""

from __future__ import annotations

from moat.lib.codec.errors import StoppedError
from moat.lib.micro import (
    Event,
    L,
    TaskGroup,
    TimeoutError,  # noqa:A004
    every_ms,
    retry_ms,
    ticks_diff,
    ticks_ms,
    wait_for_ms,
)
from moat.lib.path import Path
from moat.lib.rpc import BaseCmd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from moat.lib.rpc import Msg, SubMsgSender

    from collections.abc import Mapping


class _Send:
    # A null context that delegates its .send method to the wrapped destination
    def __init__(self, dest):
        self.dest = dest

    def send(self, *a, **kw):
        return self.dest(*a, **kw)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *tb):
        return None


class PWM(BaseCmd):
    """
    A PWM is an output pin that changes periodically.

    Parameters:
        pin: the hardware output we're controlling. Path to the write method.
        min: Minimum time between switching, milliseconds
        max: Maximum time between switching, milliseconds
        base: the maximum value for the ratio.
        init: initial value (defaults to `min`)
        so: stream_out: Flag whether to stream the pin value
        sync_low(dict): Resync settings for low transitions. The
            ``sync_low.threshold`` acts as the minimum value; output is off
            below it.
        sync_high(dict): Resync settings for high transitions. The
            ``sync_high.threshold`` acts as the maximum value; output is on
            above it.
        sync_path(Path): Path to read periodically while resync is active.
        sync_invert(bool): Invert sync_path comparison.

    The input must be in [0..base]; the output is controlled so that
    ``t_on/(t_on+t_off) = val/base``, given that ``min <= t_on,t_off <= max``
    and one of t_on and t_off are equal to `min`. (Thus when ``val=base/2``,
    both are.)

    If val is too low (or too high) such that this constraint can no longer
    be satisfied, the output is turned off (or on) permanently.

    Resync uses ``sync_low``/``sync_high`` to clamp the PWM input when
    transitioning back into range. When the input crosses above
    ``sync_low.threshold``, the PWM is set to at least ``sync_low.input`` and
    resync begins. When the input drops below ``sync_high.threshold``, the PWM
    is set to at most ``sync_high.input``. Resync ends when ``t_sync``
    expires, when the input returns out of range, or when ``t_sync`` is unset
    and the input passes the resync input. While resync is active, ``sync_path``
    can suspend resync based on ``sync_low.bound``/``sync_high.bound``.
    """

    t_last = 0
    t_on: int = 0
    t_off: int = 0
    is_on: bool = False

    value: float = 0  # must be in range 0..base
    init: float = 0  # initial value
    min: int = 500  # milliseconds
    max: int = 100000  # milliseconds
    base: float = 1000  # max for value
    evt: Event
    ps: Msg  # Data stream to the pin
    force: float | None = None

    sync_low: dict
    sync_high: dict
    sync_invert: bool = False
    sync_path: SubMsgSender | None = None

    # Sync/resync state tracking
    _out_mode: str | None = None
    _sync_active: str | None = None
    _sync_left: int | None = None
    _sync_suspended: bool = False
    _sync_check_ms: int | None = None
    _resync_scope = None  # anyio.CancelScope | None — cancel scope of resync task

    _d_threshold = dict(
        _d="describes behavior at min/max boundaries",
        bound="float:suspend resync if @sync_path exceeds this",
        t_sync="int:max resync time (s, optional)",
        t_check="int:interval for value check (s, default 10)",
        threshold="float:threshold to start resync, off below/on above",
        input="float:PWM value during resync",
    )

    doc = dict(
        _c=dict(
            _d="Slow PWM",
            pin="path:output cmd",
            max="int:max T(100000,ms)",
            min="int:min T(500,ms)",
            sync_low=_d_threshold,
            sync_high=_d_threshold,
            sync_path="path:value to check",
            sync_invert="bool:checked value inverted WRT input?",
            base="float:input range 0...(1000)",
            init="float:initial value",
            so="bool:stream to pin? (no)",
        )
    )

    def __init__(self, cfg):
        super().__init__(cfg)
        if not isinstance(cfg.get("pin", None), (tuple, list, Path)):
            raise ValueError(f"Pin not set {cfg}")  # noqa:TRY004
        self._load()
        self.evt = Event()

    def _load_sync(self, name: str) -> dict:
        cfg = self.cfg.get(name) or {}
        return dict(
            threshold=cfg.get("threshold"),
            input=cfg.get("input"),
            t_sync=cfg.get("t_sync"),
            t_check=cfg.get("t_check", 10),
            bound=cfg.get("bound"),
        )

    def _load(self):
        cfg = self.cfg
        self.min = cfg.get("min", self.min)
        self.max = cfg.get("max", self.max)
        self.base = cfg.get("base", self.base)
        self.so = self.cfg.get("so", False)
        self.sync_invert = cfg.get("sync_invert", False)
        self.sync_low = self._load_sync("sync_low")
        self.sync_high = self._load_sync("sync_high")

    def _load_sync_path(self):
        # Calling `await self.sync_path()` returns the float to check
        # cfg.sync_low.bound or cfg.sync_high.bound against.
        # Must be called after the command is attached (root is available).
        if "sync_path" not in self.cfg:
            self.sync_path = None
            return
        sp = self.root.sub_at(self.cfg.sync_path)
        if not hasattr(sp, "cmd"):
            raise TypeError(sp)
        self.sync_path = sp

    async def reload(self):
        "reload from config"
        self._load()
        self._load_sync_path()
        await super().reload()

    async def setup(self):  # noqa:D102
        await super().setup()
        self.pin = self.root.sub_at(self.cfg["pin"], cmd=not self.so)
        if await self.pin.rdy_():
            raise StoppedError("pin")
        self._load_sync_path()
        self.set_times(self.cfg.get("init", self.min))

    async def task(self):  # noqa:D102
        async with TaskGroup() as self._tg, \
                _Send(self.pin) if not self.so else self.pin.stream_out() as self.ps:  # fmt:skip
            try:
                if L:
                    self.set_ready()
                self.t_last = ticks_ms()
                self.is_on = False
                await self.ps.send(False)

                while True:
                    await self._maybe_start_resync()
                    dly = await self._measure(ticks_ms())

                    # Delay for @dly milliseconds, or until the event is set.
                    if dly is None:
                        await self.evt.wait()
                        self.evt = Event()
                    else:
                        try:
                            await wait_for_ms(dly, self.evt.wait)
                        except TimeoutError:
                            pass
                        else:
                            self.evt = Event()

            finally:
                await self.ps.send(False)

    def _sync_cfg(self, mode: str) -> dict:
        return self.sync_low if mode == "low" else self.sync_high

    def _start_resync(self, mode: str) -> None:
        cfg = self._sync_cfg(mode)
        if cfg.get("input") is None:
            return
        if self._resync_scope is not None:
            self._resync_scope.cancel()
            self._resync_scope = None
        self._sync_active = mode
        t_sync = cfg.get("t_sync")
        self._sync_left = None if t_sync is None else int(t_sync * 1000)
        self._sync_suspended = False
        if self.sync_path is not None and cfg.get("bound") is not None:
            t_check = cfg.get("t_check", 10)
            self._sync_check_ms = int(t_check * 1000)
        else:
            self._sync_check_ms = None

    def _end_resync(self) -> None:
        if self._resync_scope is not None:
            self._resync_scope.cancel()
            self._resync_scope = None
        self._sync_active = None
        self._sync_left = None
        self._sync_suspended = False
        self._sync_check_ms = None

    def _select_out_mode(self, val: float) -> str | None:
        low_threshold = self.sync_low.get("threshold")
        if low_threshold is not None and val <= low_threshold:
            return "low"
        high_threshold = self.sync_high.get("threshold")
        if high_threshold is not None and val >= high_threshold:
            return "high"
        return None

    def _sync_value(self, val: float) -> float:
        if self._sync_active == "low":
            resync_input = self.sync_low.get("input")
            if resync_input is not None:
                return max(val, resync_input)
        elif self._sync_active == "high":
            resync_input = self.sync_high.get("input")
            if resync_input is not None:
                return min(val, resync_input)
        return val

    def _apply_value(self, val: float) -> None:
        if self._out_mode == "low":
            t_on, t_off = (0, self.max)
        elif self._out_mode == "high":
            t_on, t_off = (self.max, 0)
        else:
            eff_val = val
            if self._sync_active is not None and not self._sync_suspended:
                eff_val = self._sync_value(val)
            t_on, t_off = self.calc_times(eff_val)

        self.t_on = t_on
        self.t_off = t_off

        td = ticks_diff(ticks_ms(), self.t_last)
        if td >= (t_on if self.is_on else t_off):
            self.evt.set()

    async def _resync_countdown(self, t_ms: int) -> None:
        """
        Separate task that counts down ``t_ms`` of non-suspended time and
        then ends the active resync.
        """
        t_check = self._sync_check_ms
        # Choose interval and optional path call.
        interval = t_check if t_check is not None else t_ms
        p = self.sync_path if t_check is not None else None

        remaining = t_ms
        last = ticks_ms()
        async for _value in every_ms(interval):
            now = ticks_ms()
            dt = ticks_diff(now, last)
            last = now

            if p is not None and self._sync_active is not None:
                cfg = self._sync_cfg(self._sync_active)
                bound = cfg.get("bound")
                if bound is not None:
                    sync_value = await retry_ms(0, 10, p, _exc=ValueError)
                    assert sync_value is not None
                    prev = self._sync_suspended
                    cond = (
                        sync_value < bound
                        if self.sync_invert == (self._sync_active == "low")
                        else sync_value > bound
                    )
                    self._sync_suspended = cond
                    if prev != cond:
                        self._apply_value(self.val)

            if not self._sync_suspended:
                remaining -= dt
            if remaining <= 0:
                break

        # Clear scope before _end_resync so it does not try to cancel itself.
        self._resync_scope = None
        self._end_resync()
        self._apply_value(self.val)
        self.evt.set()

    async def _maybe_start_resync(self) -> None:
        """
        Launch ``_resync_countdown`` if resync is active with a timer and no
        task is already running.

        Called at the top of the ``task()`` main loop so that resync tasks
        started by ``set_times`` (sync) are picked up on the next scheduler
        turn.
        """
        if (
            self._sync_active is not None
            and self._sync_left is not None
            and self._resync_scope is None
        ):
            sync_left = self._sync_left

            async def _run_resync() -> None:
                await self._resync_countdown(sync_left)

            self._resync_scope = await self._tg.spawn(_run_resync, _name="Resync")

    async def _measure(self, now: int) -> int | None:
        """
        Check whether it's time to switch.

        Returns: delay until the next switch, or ``None`` for
        "until the value is changed".
        """
        td = ticks_diff(now, self.t_last)

        async def _sw(state: bool) -> int | None:
            nonlocal now

            if self.is_on != state:
                await self.ps.send(state)
                self.is_on = state
                self.t_last = now
            if state:
                return self.t_on if self.t_off else None
            else:
                return self.t_off if self.t_on else None

        dly = None
        if self.t_on == 0:
            # switch off
            dly = await _sw(False) if td >= self.min else int(self.min - td)
        elif self.t_off == 0:
            dly = await _sw(True) if td >= self.min else int(self.min - td)

        elif self.is_on:
            dly = await _sw(False) if td >= self.t_on else int(self.t_on - td)
        else:
            dly = await _sw(True) if td >= self.t_off else int(self.t_off - td)
        return dly

    def calc_times(self, val: float) -> tuple[int, int]:
        """
        Calculate the t_on/t_off tuple so that
        ``t_on/(t_on+t_off) == val/base`` and ``min <= t_{on,off} <= max``.

        If that ratio falls below ``min/(min+max)``, switch off entirely
        (t_on=0). Likewise, turn on when the ratio is too high.
        """
        rev = False

        a = self.min
        b = self.max
        base = self.base

        if val * 2 > base:
            rev = True
            val = base - val

        # a/(a+b) is the minimum ratio. Below half that we switch off,
        # i.e. val/base < a/(a+b)/2 -- reordered to avoid division.
        if val * (a + b) * 2 < a * base:
            a = 0
        else:
            # a/(a+b) == val/base; solve for b
            # the test above prevents val from being zero
            r = int(a * (base - val) / val)  # times are integer msec
            b = min(b, r)

        return (b, a) if rev else (a, b)

    def set_times(self, val: float, force: bool = False):
        """
        Change the on/off ratio to approximately ``val/base``.
        """
        if val is None:
            if not force:
                raise ValueError(val)
            self.force = None
            return

        if val < 0 or val > self.base:
            raise ValueError(val, self.base)
        if force:
            self.force = val
        else:
            self.value = val

        prev_out = self._out_mode
        new_out = self._select_out_mode(val)
        self._out_mode = new_out

        if new_out is not None:
            if self._sync_active is not None:
                self._end_resync()
        elif prev_out is not None and self._sync_active is None:
            self._start_resync(prev_out)

        if self._sync_active == "low" and self._sync_left is None:
            resync_input = self.sync_low.get("input")
            if resync_input is not None and val >= resync_input:
                self._end_resync()
        elif self._sync_active == "high" and self._sync_left is None:
            resync_input = self.sync_high.get("input")
            if resync_input is not None and val <= resync_input:
                self._end_resync()

        self._apply_value(val)

    doc_w = dict(
        _d="change",
        _0="float|None:new value",
        f="bool:forced value",
        _i=dict(_0="float:new value"),
    )

    async def stream_w(self, msg: Msg):
        "change ratio"
        if msg.can_stream:
            async with msg.stream_in() as md:
                async for m in md:
                    self.set_times(m[0], msg.get("f", False))
        else:
            self.set_times(msg[0], msg.get("f", False))

    doc_s = dict(
        _d="read state",
        _r=dict(
            on="int:t_on",
            off="int:t_off",
            p="bool:state",
            t="int:time until next change",
        ),
    )

    @property
    def val(self) -> float:
        """
        The current value.
        """
        if self.force is not None:
            return self.force
        return self.value

    async def cmd_r(self) -> float:
        """
        Returns the current value.
        """
        return self.val

    async def cmd_s(self) -> Mapping:
        """
        Returns the current state.
        """
        now = ticks_ms()
        res: dict[str, object] = dict(
            on=self.t_on,
            off=self.t_off,
            p=self.is_on,
            val=self.value,
        )
        if self.force is not None:
            res["force"] = self.force
        if self._sync_active is not None:
            res["resync"] = dict(
                mode=self._sync_active,
                suspended=self._sync_suspended,
            )
        elif self._out_mode is not None:
            res["out_mode"] = self._out_mode
        elif self.t_on and self.t_off:
            res["t"] = (self.t_on if self.is_on else self.t_off) - ticks_diff(now, self.t_last)
        return res



"""
Basic infrastructure to run an RPC system via an unreliable,
possibly-reordering, and/or stream-based transport

We have a stack of classes. At the top there's the adapting App, a subclass
of CmdMsg. Linked to it via a chain of *Msg…Buf modules there's a
*Buf adapter that affords an external interface.

Everything is fully asynchronous. Each class is an async context manager,
responsible for managing the contexts of its sub-app(s) / linked stream(s).

Incoming messages are handled by the child's "dispatch" method. They
are expected to be fully asynchronous, i.e. a "run" method that calls
"dispatch" must use a separate task to do so.

Outgoing messages are handled by the parent's "send" method. Send calls
return when the data has been sent, implying that sending on an
unreliable transport will wait for the message to be confirmed. Sending
may fail.
"""

from __future__ import annotations

import sys

from moat.util import enc_part, get_part, import_, wait_complain
from moat.lib.micro import (
    AC_use,
    Event,
    L,
    Lock,
    ModuleNotFoundError,  # noqa:A004
    ObjSequence,
    TaskGroup,
    idle,
)
from moat.lib.path import Path
from moat.lib.rpc import MsgHandler, MsgSender
from moat.lib.stream import Base

from collections.abc import Mapping

from typing import TYPE_CHECKING, cast  # isort:skip

if TYPE_CHECKING:
    from contextlib import AbstractAsyncContextManager

    from moat.lib.micro import _TaskGroupProto
    from moat.lib.rpc import BaseSuperCmd
    from moat.lib.rpc.base import MsgRoot

    from typing import Literal, Protocol

    class _Cancelable(Protocol):
        def cancel(self) -> None: ...


__all__ = ["BaseCmd", "LoadCmd", "LockBaseCmd", "RootCmd", "add_app_prefix", "load_app"]

APP: list[str] = ["app"] if sys.implementation.name == "micropython" else ["moat.lib.rpc.app"]


def add_app_prefix(prefix: str) -> None:
    """
    Register a module prefix for dynamic command loading.

    The prefix should end in `.app`, by convention.
    """
    APP.append(prefix)


def load_app(name: str):
    """
    Load an app class/object from the registered app prefixes.
    """
    err = None
    for app in APP:
        try:
            return import_(f"{app}.{name}", 1)
        except (AttributeError, ModuleNotFoundError) as exc:
            err = exc
    if err is not None:
        raise err
    raise ImportError(name)


class BaseCmd(MsgHandler):
    """
    Base class of a Command handler.
    """

    root: RootCmd = None  # type:ignore[invalid-assignment]
    real_root: RootCmd
    # real_root is set when `root` is a non-settable property.
    # This is used for root hijackers like `log.Cmd` so that they can
    # capture outgoing requests.

    _parent: BaseSuperCmd | None = None
    _name: str | None = None
    _ts = None
    _rl_ok = None  # result of last reload
    p_task: _Cancelable | Literal[False] | None = None  # managed by parent. Do not touch.

    if L:
        _starting: Event | None = None
        _ready: Event | None = None
    _stopped: Event | None = None

    def __init__(self, cfg):
        cfg._moat_cmd = self  # noqa:SLF001
        super().__init__(cfg)
        # self.cfg = cfg
        self.init_events()

    def init_events(self):
        "Setup events for startup, stopping, and readiness"
        if self._stopped is None:
            if L:
                self._starting = Event()
                self._ready = Event()
            self._stopped = Event()

    def __repr__(self):
        try:
            return f"<{self.cfg_name}: {self.path} {(id(self) >> 4) & 0xFFF:03x}>"
        except AttributeError:
            return f"<{self.cfg_name}: ?path {(id(self) >> 4) & 0xFFF:03x}>"

    async def setup(self):
        """
        Start up this command.

        Call this first when overriding.
        """
        await super().setup()
        await AC_use(self, self.set_stopped)
        if self._stopped is None:
            self.init_events()
        elif L:
            if self._starting is None or self._starting.is_set():
                raise RuntimeError("DupStartA")

            self._starting.set()
            self._starting = None

    async def reload(self):
        """
        Reload from updated config.

        The default does nothing, which is probably the wrong thing to do.
        """

    def set_stopped(self):
        """
        The command has ended.
        """
        if self._stopped is None:
            return
        self._stopped.set()
        self._stopped = None

        if L:
            if self._starting is not None:
                self._starting.set()
                self._starting = None

            if self._ready is not None:
                self._ready.set()
                self._ready = None

    async def run(self):
        """
        Task handler to run this command.

        If you need a subtask, override ``task``.
        """
        async with self:
            await self.task()

    async def task(self):
        """
        The app's task. Runs after ``setup`` has completed. By default does nothing.

        If you override this, you're responsible for eventually calling ``set_ready``.
        Otherwise the MoaT system will stall!
        """
        if L:
            self.set_ready()
        await idle()

    if L:

        def set_ready(self):
            """
            This command is now ready.

            Called internally only!
            """
            self.cfg.pop("_cmd", None)
            if self._starting is not None:
                raise RuntimeError(f"Ready w/o start {self!r}")
                # self._starting.set()
                # self._starting = None
            if self._ready is not None:
                self._ready.set()
                self._ready = None

    async def stop(self):
        "Stop this subcommand"
        p_task = self.p_task
        if p_task is None:
            self.p_task = False
            return  # starting up or not running
        if p_task is False:
            return
        p_task.cancel()
        await wait_complain(f"Stop {self.path}", 250, self.wait_stopped)

    cmd_stp_ = stop

    if L:

        async def wait_started(self):
            """
            Wait for the command's setup procedure to commence.
            """
            if self._starting is None:
                return
            await wait_complain(f"Starting {self.path}", 250, self._starting.wait)

        async def wait_ready(self, wait=True) -> bool | None:
            """
            Check if the command is ready.

            Returns True if it is stopped, False if it is already running, and
            None if the command has (or would have, if @wait is False) waited
            for it to become ready.
            """
            if self._stopped is None:
                return True
            if self._ready is None:
                return False
            if wait:
                await wait_complain(f"Rdy {self.path}", 250, self._ready.wait)
            return None

    async def wait_stopped(self):
        "wait until this is no longer running"
        if self._stopped is not None:
            await self._stopped.wait()

    doc_stq_ = dict(_d="wait until no longer running")

    cmd_stq_ = wait_stopped

    @property
    def cfg_name(self) -> str:
        "Return a human-readable name"
        try:
            return self.cfg["app"]
        except KeyError:
            return super().cfg_name

    @property
    def path(self):
        "calculate the path to myself"
        # XXX cache it?
        if self._parent is None:
            raise RuntimeError("Not attached")
        if self._name is None:
            return self._parent.path
        return self._parent.path / self._name

    doc_cfg_ = dict(_d="config", _r="parts")

    async def cmd_cfg_(self, p=()):
        """
        Read this item's config.

        Writing is intentionally not supported.
        """
        return enc_part(get_part(self.cfg, p))

    def attached(self, parent: BaseSuperCmd, name: str | None):
        "Tell this Cmd it's attached under this parent, with this name."
        if self._parent is not None:
            raise RuntimeError(f"already {'.'.join(self.path)}")
        self._parent = parent
        self._name = name

        root = parent.root
        if root is None:
            raise RuntimeError("No root set")
        try:
            self.root = root
        except AttributeError:
            self.real_root = root


class LockBaseCmd(BaseCmd):
    """
    A BaseCmd that blocks concurrent requests.
    """

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._lock = Lock()

    async def handle(self, *a, **kw):
        "strictly serial"
        async with self._lock:
            return await super().handle(*a, **kw)


class LoadCmd(BaseCmd):
    """
    A command loaded from disk.
    """

    def __new__(cls, cfg, **kw):
        if not cfg.get("running", True):
            return None

        def imp(name):
            if name == "dir":
                name = "sub.Dir"
            return load_app(name)

        return imp(cfg["app"])(cfg, **kw)

    def __init__(self, cfg, **kw):
        pass


class RootCmd(Base):
    """
    This is the system's root dispatcher.

    It delegates most (if not all) to its app.
    """

    tg: object | None = None

    def __init__(self, cfg, i=None):
        self.i = i
        self._sender = MsgSender(cast("MsgRoot", self))
        self.cfg = cfg
        if not isinstance(cfg["app"], str):
            cfg = cfg["app"]
        self.app = LoadCmd(cfg)
        self.app.attached(cast("BaseSuperCmd", self), None)
        self._updates = {}

    def __repr__(self):
        if self.__class__ is RootCmd:
            return f"<{self.__class__.__name__}>"
        return f"<Root:{self.__class__.__name__}>"

    async def setup(self):
        await super().setup()
        tg = await AC_use(self, cast("AbstractAsyncContextManager[object]", TaskGroup()))
        self.tg = tg
        tg = cast("_TaskGroupProto", tg)
        await AC_use(self, tg.cancel)
        tg.start_soon(self.app.run)
        tg.start_soon(self.task)
        if L:
            await self.wait_ready()

    async def teardown(self):
        self.tg = None
        await super().teardown()

    async def task(self):
        """
        Background task. May be overridden.
        """
        pass

    async def run(self):
        """
        This method should not be called. Roots use context management for
        (sub)task control.
        """
        raise NotImplementedError

    if L:

        async def wait_ready(self):
            return await self.app.wait_ready()

    def __getattr__(self, k):
        if k[0] == "_":
            raise AttributeError(k)
        return getattr(self.app, k)

    @property
    def root(self) -> RootCmd:
        "root dispatcher"
        return self

    @property
    def sender(self) -> MsgSender:
        "sender adapter"
        return self._sender

    @property
    def path(self):
        "root path"
        return Path()

    @property
    def cmd(self):
        "root command sender"
        return self._sender.cmd

    @property
    def sub_at(self):
        "root subcommand resolver"
        return self._sender.sub_at

    @property
    def cfg_at(self):
        "config subcommand resolver"
        return self._sender.cfg_at

    async def reload(self):
        "Reload me."
        await self.app.reload()

        upd, self._updates = self._updates, {}

        def _upd(val):
            if hasattr(val, "updated_"):
                val.updated_()
            if isinstance(val, ObjSequence):
                for v in val:
                    _upd(v)
            elif isinstance(val, Mapping):
                for v in val.values():
                    _upd(v)

        _upd(upd)

    def cfg_updated(self, cfg):
        "Mark TODO for update"
        self._updates[id(cfg)] = cfg

    def cfg_reloaded(self, cfg):
        "Remove TODO for update"
        self._updates.pop(id(cfg), None)



"""
Test support
"""

from __future__ import annotations

import anyio
from contextlib import asynccontextmanager
from contextvars import ContextVar
from pathlib import Path
from random import random

from moat.util import attrdict, combine_dict, ctx_as, yload
from moat.lib.rpc import RootCmd
from moat.lib.stream import BaseBlk, BaseBuf, BaseMsg

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from types import CoroutineType

    from moat.lib.micro import Event
    from moat.lib.stream.base import Buffer, MutBuffer

    from typing import Any, Protocol

    class _LoopLinkProto(Protocol):
        def xs(self, *, m: Any) -> CoroutineType[Any, Any, None]: ...

        def xr(self) -> CoroutineType[Any, Any, Any]: ...

        def xsb(self, *, m: Buffer | bytes) -> CoroutineType[Any, Any, None]: ...

        def xrb(self) -> CoroutineType[Any, Any, Buffer | bytes]: ...

        def xwr(self, *, b: Buffer) -> CoroutineType[Any, Any, None]: ...

        def xrd(self, *, n: int) -> CoroutineType[Any, Any, Buffer | bytes]: ...

        def xcwr(self, *, b: Buffer) -> CoroutineType[Any, Any, None]: ...

        def xcrd(self, *, n: int) -> CoroutineType[Any, Any, Buffer | bytes]: ...


temp_dir = ContextVar("temp_dir")


@asynccontextmanager
async def rpc_stack(temp: Path, cfg: attrdict | str, cfg2: dict | None = None, **kw):
    """
    Creates a multiplexer.
    """
    if isinstance(cfg, str):
        if "\n" in cfg:
            cfg = yload(cfg, attr=True)
        else:
            with (Path("tests") / "cfg" / (cfg + ".cfg")).open("r") as cff:
                cfg = yload(cff, attr=True)

    if cfg2 is not None:
        cfg = combine_dict(cfg2, cfg, cls=attrdict)

    async with ctx_as(temp_dir, temp):
        if isinstance(cfg.app, str):
            cfg = attrdict(app=cfg)
        if "rtc" in cfg:
            from moat.micro.rtc import RTC  # noqa:PLC0415

            RTC.init(cfg["rtc"])
        stack = RootCmd(cfg, **kw)
        async with stack:
            yield stack


class Loopback(BaseMsg, BaseBuf, BaseBlk):
    """
    A simple loopback object.

    The write queue is created locally, the read queue is taken from the
    "other side".

    This object can be self-linked.
    """

    # pylint:disable=abstract-method

    _link: Event | Loopback | None = None
    _buf: MutBuffer

    def __init__(self, qlen=0, loss=0):
        super().__init__({})
        assert 0 <= loss < 1
        self.q_wr, self.q_rd = anyio.create_memory_object_stream(qlen)
        self.loss = loss

    async def setup(self):
        if self._link is None:
            raise RuntimeError("You must link before setup!")
        elif isinstance(self._link, anyio.Event):
            await self._link.wait()

    def link(self, other: Loopback | anyio.Event):
        """Tell this loopback to read from some other loopback."""
        evt, self._link = self._link, other
        if isinstance(evt, anyio.Event):
            evt.set()

    async def send(self, m, _loss=True):  # pylint:disable=arguments-differ
        """Send data."""
        if self._link is None:
            raise anyio.BrokenResourceError(self)
        if _loss and random() < self.loss:
            return
        try:
            await self.q_wr.send(m)
        except (
            anyio.ClosedResourceError,
            anyio.BrokenResourceError,
            anyio.EndOfStream,
        ) as exc:
            raise EOFError from exc

    snd = send

    async def recv(self):  # pylint:disable=arguments-differ
        if (link := self._link) is None:
            raise anyio.BrokenResourceError(self)
        try:
            return await cast(Loopback, link).q_rd.receive()
        except (
            anyio.ClosedResourceError,
            anyio.BrokenResourceError,
            anyio.EndOfStream,
        ):
            raise EOFError from None

    rcv = recv

    async def rd(self, buf: MutBuffer) -> int:
        while True:
            try:
                b = self._buf
            except AttributeError:
                pass
            else:
                if b:
                    n = min(len(b), len(buf))
                    buf[0:n] = b[0:n]
                    self._buf = b[n:]
                    return n
            self._buf = await self.recv()

    async def wr(self, data: Buffer) -> int:
        n = len(data)
        if self.loss:
            b = bytearray(data)
            loss = 1 - (1 - self.loss) ** (1 / len(b) / 2)
            # '1-loss' is the chance of not killing each single byte
            # that's required to not kill a message of size len(b)
            # given two chances of mangling each byte

            n = 0
            while n < len(b):
                if random() < loss:
                    del b[n]
                else:
                    while random() < loss:
                        b[n] = b[n] ^ (1 << int(8 * random()))
                    n += 1
        else:
            b = bytes(data)
        await self.send(b, _loss=False)
        return n

    async def teardown(self):
        await self.q_wr.aclose()
        link = self._link
        if isinstance(link, Loopback) and link is not self:
            await link.q_rd.aclose()
        await super().teardown()


class LoopBBM(BaseMsg, BaseBuf, BaseBlk):
    """
    A loopback BBM. It talks to a remote LoopLink.

    This BBM is not a command, thus it cannot be linked to.

    The remote LoopLink must have the appropriate buffers,
    i.e. `usage: mM` for messages, etc.
    """

    # pylint:disable=abstract-method

    _link: _LoopLinkProto

    async def setup(self):
        p = self.cfg["path"]
        if isinstance(p, str):
            raise TypeError(f"Need a path, not {p!r}")
        self._link = self.cfg._moat_cmd.root.sub_at(p)  # noqa:SLF001

    def send(self, m: Any) -> CoroutineType[Any, Any, None]:
        """Send message data."""
        return self._link.xs(m=m)

    def recv(self) -> CoroutineType[Any, Any, Any]:
        """Read message data."""
        return self._link.xr()

    def snd(self, m: Buffer | bytes) -> CoroutineType[Any, Any, None]:
        """Send block data."""
        return self._link.xsb(m=m)

    def rcv(self) -> CoroutineType[Any, Any, Buffer | bytes]:
        """Read block data."""
        return self._link.xrb()

    async def wr(self, data: Buffer) -> int:
        """Send bytes."""
        await self._link.xwr(b=data)
        return len(data)

    async def rd(self, buf: MutBuffer) -> int:
        """Read bytes."""
        r = await self._link.xrd(n=len(buf))
        n = len(r)
        buf[:n] = r
        return n

    def cwr(self, b: Buffer) -> CoroutineType[Any, Any, None]:
        """Send bytes."""
        return self._link.xcwr(b=b)

    async def crd(self, b: bytearray) -> int:
        """Read bytes."""
        r = await self._link.xcrd(n=len(b))
        n = len(r)
        b[:n] = r
        return n


class Root(RootCmd):
    "an empty root for testing"

    def __init__(self):
        super().__init__({})



"""
FUSE operations for MoaT-micro-FS
"""

from __future__ import annotations

from moat.util import ExpAttrError, ExpKeyError
from moat.lib.proxy import as_proxy

try:
    from moat.lib.rpc import RemoteError
except ImportError:

    class RemoteError(RuntimeError):  # noqa: D101
        pass


@as_proxy("_rErrS")
class SilentRemoteError(RemoteError):  # type: ignore[misc]  # RemoteError may be from two different sources
    """
    Forwarded error from a remote system.

    Unlike `RemoteError`, this should not trogger a stack dump.
    """


@as_proxy("_NPErr")
class NoPathError(KeyError):
    """An error that marks a nonexisting path"""

    def __str__(self):
        from moat.lib.path import Path  # noqa: PLC0415

        return (
            f"‹NoPath {self.args[0]} {Path.build(self.args[1])}"
            f"{' ' + ' '.join(str(x) for x in self.args[2:]) if len(self.args) > 2 else ''}›"
        )

    def prefixed(self, path):
        "prefix the error's path with another"
        return NoPathError(path / self.args[0], *self.args[1:])


try:
    FileNotFoundError = FileNotFoundError  # noqa:A001,PLW0127
except NameError:

    class FileNotFoundError(SilentRemoteError):  # noqa:A001
        "standard exception"

        def __reduce__(self):
            return (FileNotFoundError, (self.args[0],), {})


as_proxy("_FnErr", FileNotFoundError)

try:
    FileExistsError = FileExistsError  # noqa:A001,PLW0127
except NameError:

    class FileExistsError(SilentRemoteError):  # noqa:A001
        "standard exception"

        def __reduce__(self):
            return (FileExistsError, (self.args[0],), {})


as_proxy("_FxErr", FileExistsError)

as_proxy("_KyErr", KeyError)
as_proxy("_ValErr", ValueError)
as_proxy("_EKyErr", ExpKeyError)
as_proxy("_IdxErr", IndexError)
as_proxy("_AtErr", AttributeError)
as_proxy("_EAtErr", ExpAttrError)
as_proxy("_NiErr", NotImplementedError)
as_proxy("_StpIter", StopAsyncIteration)

as_proxy("_RemErr", RemoteError)
as_proxy("_SRemErr", SilentRemoteError)


@as_proxy("_StpErr")
class StoppedError(Exception):
    "Called command/app is not running"


@as_proxy("_rErrCCl")
class ChannelClosed(RuntimeError):
    "Link closed."



"""
RTC memory backend.

Uses the hardware RTC memory for fast, synchronous access.
Data is stored as a CBOR-encoded dict tagged with CBOR_TAG_CBOR_FILEHEADER.
"""

from __future__ import annotations

from moat.util import NotGiven
from moat.lib.codec.cbor import CBOR_TAG_CBOR_FILEHEADER, Tag
from moat.lib.codec.moat_cbor import Codec as CBOR
from moat.lib.micro import log

from . import RTCBase

try:
    from machine import RTC as _RTC
except ImportError:
    _RTC = None

_codec = CBOR()


class RTC(RTCBase):
    """
    RTC memory backend.

    Uses machine.RTC().memory for hardware access.
    Data is stored as a CBOR-encoded dict tagged with CBOR_TAG_CBOR_FILEHEADER.
    Loads on init, saves on every write.
    """

    is_FS = False
    is_ASYNC = False

    def __init__(self, cfg):
        super().__init__(cfg)
        self._d = {}
        self._mem = None

        if _RTC is None:
            raise ImportError("No machine.RTC available")

        try:
            r = _RTC()
            self._mem = r.memory
        except (TypeError, AttributeError) as exc:
            raise ImportError("RTC memory not available") from exc

        self._load()

    def _load(self):
        """Load data from RTC memory."""
        if self._mem is None:
            return

        try:
            data = self._mem()
            if data and data[0:3] == b"\xd9\xd9\xf7":
                decoded = _codec.decode(data)
                if isinstance(decoded, Tag) and decoded.tag == CBOR_TAG_CBOR_FILEHEADER:
                    self._d = decoded.value
                elif isinstance(decoded, dict):
                    self._d = decoded
        except Exception as exc:
            log("RTC mem load error", err=exc)
            self._d = {}

    def _save(self):
        """Save data to RTC memory."""
        if self._mem is None:
            return
        tagged = Tag(CBOR_TAG_CBOR_FILEHEADER, self._d)
        self._mem(_codec.encode(tagged))

    def reload(self):
        """Reload data from RTC memory."""
        self._load()

    async def get(self, name: str):
        """
        Get a value from RTC memory.

        Args:
            name: The key to retrieve.
            all: Unused, for API compatibility.

        Returns:
            The stored value.

        Raises:
            KeyError: If the key doesn't exist.
        """
        if name not in self._d:
            raise KeyError(name)
        return self._d[name]

    async def set(self, name: str, data) -> bool:
        """
        Set a value in RTC memory.

        Args:
            name: The key to store.
            data: The value to store (must be CBOR-encodable).

        Returns:
            True if the value was written.
        """
        if data is NotGiven:
            if name in self._d:
                del self._d[name]
                self._save()
            return True

        self._d[name] = data
        self._save()
        return True

    async def keys(self):
        """
        Get keys for all stored data.

        Returns:
            A set of names.
        """
        return set(self._d.keys())



"""
Async front-end for the characer-based ring buffer.
"""

from __future__ import annotations

from moat.lib.micro import Event

from ._impl import RingBuffer as _RingBuf

from typing import cast


class RingBuffer(_RingBuf):
    """
    This ring buffer can hold a predetermined number of bytes.

    This is the async version.
    """

    _w_evt: Event | None = None
    _r_evt: Event | None = None

    def __init__(self, length: int):
        super().__init__(length)

    async def write(self, buf: bytes) -> int:  # type:ignore[invalid-method-override]
        """
        Adds the bytes in `buf` to the end of the buffer.
        """
        if not buf:
            return 0

        buf_len = len(buf)
        n = 0

        while True:
            nb = super().write(buf, drop=False)
            n += nb
            if self._r_evt is not None:
                self._r_evt.set()
                self._r_evt = None
            if n == buf_len:
                return buf_len
            buf = cast(bytes, memoryview(buf)[n:])
            if self._w_evt is None:
                self._w_evt = Event()
            await self._w_evt.wait()

    async def readinto(self, buf: bytearray) -> int:  # type:ignore[invalid-method-override]
        """
        Copies as many bytes as will fit (or are available, whichever is
        smaller) into the buffer and advance the read counter.
        """
        if len(buf) == 0:
            return 0
        n = super().readinto(buf)
        while n == 0:
            if self._r_evt is None:
                self._r_evt = Event()
            await self._r_evt.wait()
            n = super().readinto(buf)

        if self._w_evt is not None:
            self._w_evt.set()
            self._w_evt = None

        return n

    async def wait_avail(self) -> None:
        """
        Waits until data are available.
        """
        if self.n_avail > 0:
            return
        if self._r_evt is None:
            self._r_evt = Event()
        await self._r_evt.wait()



"""
Clone of asyncio.queue
"""

from __future__ import annotations

from asyncio import core

from collections import deque


class QueueEmpty(Exception):
    """Exception raised by get_nowait()."""


class QueueFull(Exception):
    """Exception raised by put_nowait()."""


class Queue:
    """A queue, useful for coordinating producer and consumer coroutines.

    @maxsize defaults to 99. (On MicroPython, the queue cannot be infinitely
    long.) "await put()" will block when the queue reaches maxsize, until
    an item is removed by get().

    Unlike the standard library Queue, you can reliably know this Queue's size
    with qsize(), since your single-threaded uasyncio application won't be
    interrupted between calling qsize() and doing an operation on the Queue.
    """

    def __init__(self, maxsize=99):
        self.maxsize = maxsize
        self._queue = deque((), maxsize)
        self._full = core.TaskQueue()
        self._empty = core.TaskQueue()
        self._closed_w = False

    def _get(self):
        res = self._queue.popleft()
        if self._full.peek():
            core._task_queue.push(self._full.pop())  # noqa:SLF001
        return res

    def get(self):
        """Returns generator, which can be used for getting (and removing)
        an item from a queue.

        Usage::

            item = await queue.get()
        """
        while not self._queue:
            if self._closed_w:
                raise EOFError
            self._empty.push(core.cur_task)
            core.cur_task.data = self._empty
            yield
            if self._closed_w:
                raise EOFError

        return self._get()

    def get_nowait(self):
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        if not self._queue:
            raise EOFError if self._closed_w else QueueEmpty
        return self._get()

    def _put(self, val):
        if self._empty.peek():
            core._task_queue.push(self._empty.pop())  # noqa:SLF001
        self._queue.append(val)

    def put(self, val):
        """Returns generator which can be used for putting item in a queue.

        Usage::

            await queue.put(item)
        """
        while self.maxsize and self.qsize() >= self.maxsize:
            self._full.push(core.cur_task)
            core.cur_task.data = self._full
            yield
        self._put(val)

    def put_nowait(self, val):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        if self.maxsize and self.qsize() >= self.maxsize:
            raise QueueFull
        self._put(val)

    def qsize(self):
        """Number of items in the queue."""
        return len(self._queue)

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return not self._queue

    def full(self):
        """Return True if there are maxsize items in the queue.

        Note: if the Queue was initialized with maxsize=0 (the default),
        then full() is never True.
        """
        if self.maxsize <= 0:
            return False
        else:
            return self.qsize() >= self.maxsize

    def close_sender(self):
        """
        Close the send part of the queue.

        Readers will get an EOFError after they drained it.
        """
        self._closed_w = True
        while self._empty.peek():
            core._task_queue.push(self._empty.pop())  # noqa:SLF001



"""
A hacked-up Mersenne Twister.
"""

from __future__ import annotations


class Random:
    "simplified random number generator"

    def __init__(self, c_seed=0):
        # MT19937 except with a smaller period
        (self.w, self.n, self.m, self.r) = (32, 624, 397, 31)
        self.n = 32  ## reduce memory usage
        self.a = 0x9908B0DF
        (self.u, self.d) = (11, 0xFFFFFFFF)
        (self.s, self.b) = (7, 0x9D2C5680)
        (self.t, self.c) = (15, 0xEFC60000)
        self.l = 18
        self.f = 1812433253
        # make a arry to store the state of the generator
        self.MT = [0 for i in range(self.n)]
        self.index = self.n + 1
        self.lower_mask = 0x7FFFFFFF
        self.upper_mask = 0x80000000
        # inital the seed
        self.c_seed = c_seed
        self.seed(c_seed)

    def seed(self, num):
        """initialize the generator from a seed"""
        self.MT[0] = num
        self.index = self.n
        for i in range(1, self.n):
            temp = self.f * (self.MT[i - 1] ^ (self.MT[i - 1] >> (self.w - 2))) + i
            self.MT[i] = temp & 0xFFFFFFFF

    def twist(self):
        """Generate the next n values from the series x_i"""
        for i in range(self.n):
            x = (self.MT[i] & self.upper_mask) + (self.MT[(i + 1) % self.n] & self.lower_mask)
            xA = x >> 1
            if (x % 2) != 0:
                xA = xA ^ self.a
            self.MT[i] = self.MT[(i + self.m) % self.n] ^ xA
        self.index = 0

    def extract_number(self):
        """Extract a tempered value based on MT[index]
        calling twist() every n numbers
        """
        if self.index >= self.n:
            self.twist()

        y = self.MT[self.index]
        y = y ^ ((y >> self.u) & self.d)
        y = y ^ ((y << self.s) & self.b)
        y = y ^ ((y << self.t) & self.c)
        y = y ^ (y >> self.l)

        self.index += 1
        return y & 0xFFFFFFFF

    def random(self):
        """return uniform ditribution in [0,1)"""
        # a = (self.extract_number() / 10**8) % 1
        # return float('%.08f' % a)
        return self.extract_number() / 4294967296  # which is 2**w

    def randint(self, a, b):
        """return random int in [a,b)"""
        n = self.random()
        return int(n / (1 / (b - a)) + a)



"""
Main entry point for satellites.

Called from the root "main.py".
"""

from __future__ import annotations

import os
import sys

import machine

import moat.micro.console as cons
from moat.util import attrdict, merge, to_attrdict
from moat.lib.codec.moat_cbor import Codec as CBOR
from moat.lib.micro import AC_use, L, TaskGroup, sleep_ms
from moat.lib.rpc import MsgSender

WDT = None


def wr_exc(exc):
    "write exception and backtrace to stderr, and maybe flash"
    if "moat.err" not in os.listdir():
        with open("moat.err", "w") as f:
            print("Error:", repr(exc), file=f)
            sys.print_exception(exc, f)

    print("Error:", repr(exc), file=sys.stderr)
    sys.print_exception(exc, sys.stderr)


def main(cfg: str | dict, i: attrdict, fake_end=False) -> None:
    """
    The MoaT.micro satellite's main entry point.

    Args:
        cfg: the configuration. If this is a string, it's the name of the
             config file, with the data stored in CBOR.

        fake_end: if set, sends a fake MicroPython prompt to trick the setup code into
                  thinking that the current command has concluded. This way
                  it can cleanly terminate the setup phase and start the
                  local dispatcher.
    """
    if isinstance(cfg, str):
        with open(cfg, "rb") as f:
            cfg = CBOR().decode(f.read())
    if type(cfg) is dict:
        cfg = to_attrdict(cfg)

    from moat.micro.rtc import RTC  # noqa:PLC0415

    RTC.init(cfg.get("rtc", {}))

    # Update config from RTC memory, if present
    if not i["fb"]:
        try:
            rcfg = RTC.get_sync("cfg")
        except KeyError:
            pass
        else:
            merge(cfg, rcfg)

    # Start the hardware watchdog timer early
    for v in cfg.values():
        if not isinstance(v, dict):
            continue
        if v.get("app", "") != "wdt.Cmd":
            continue
        if v.get("running", True) and v.get("hw", False):
            machine.WDT(v.get("id", 0), v.get("t", 5000))
        break

    def cfg_network(n):
        import time  # noqa: PLC0415

        import network  # noqa: PLC0415

        network.hostname(n["name"])
        if "country" in n:
            network.country(n["country"])
        if "ap" in n:
            wlan = network.WLAN(network.STA_IF)  # create station interface
            wlan.active(True)
            if "addr" in n:
                nm = n.get("netmask", 24)
                if isinstance(nm, int):
                    ff = (1 << 32) - 1
                    nm = (ff << (32 - nm)) & ff
                    nm = f"{(nm >> 24) & 0xFF}.{(nm >> 16) & 0xFF}.{(nm >> 8) & 0xFF}.{nm & 0xFF}"
                wlan.ifconfig((n["addr"], n["netmask"], n["router"], n["dns"]))
            wlan.connect(n["ap"], n.get("pwd", ""))  # connect to an AP
        else:
            wlan = network.WLAN(network.AP_IF)  # create a station interface

        n = 0
        if wlan.isconnected():
            return
        print("WLAN", end="", file=sys.stderr)
        while not wlan.isconnected():
            if n > 300:
                print(" - no connection", file=sys.stderr)
                raise RuntimeError("no network link")
            n += 1
            time.sleep_ms(100)
            print(".", end="", file=sys.stderr)
        print(" -", wlan.ifconfig()[0], file=sys.stderr)

    if "net" in cfg and cfg["net"].get("name", None) is not None:
        cfg_network(cfg["net"])

    async def _main():
        import sys  # noqa: PLC0415

        from moat.lib.rpc import RootCmd  # noqa: PLC0415

        cons.main = m = cons.Main(wr_exc)
        dsp = await AC_use(m, RootCmd(cfg, i=i))
        m.tg = await AC_use(m, TaskGroup())
        m.root = MsgSender(dsp)

        m.main_task = await m.tg.spawn(dsp.task)
        if L:
            await dsp.wait_ready()
        else:
            await sleep_ms(1000)
        if fake_end:
            sys.stdout.write("OK\x04\x04>")
        else:
            print("MoaT is up.", file=sys.stderr)

        try:
            await m.wait()
        except BaseException as exc:
            await m.die_(exc)
            print("MoaT has terminated.", file=sys.stderr)
        else:
            if m.console is None:
                print("MoaT is down.", file=sys.stderr)
            else:
                print("MoaT is in the background.", file=sys.stderr)

    from asyncio import create_task, run_until_complete  # noqa: PLC0415

    task = create_task(_main())
    run_until_complete(task)



"""
Basic BMS classes
"""

from __future__ import annotations

from moat.util import attrdict, val2pos
from moat.lib.micro import (
    Event,
    TaskGroup,
    TimeoutError,  # noqa:A004
    log,
    sleep_ms,
    ticks_diff,
    ticks_ms,
    wait_for_ms,
)
from moat.lib.proxy import as_proxy
from moat.lib.rpc import ArrayCmd, BaseCmd
from moat.lib.rpc.alert import Alert
from moat.micro.rtc import state as rtc_state

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.lib.rpc import SubMsgSender
    from moat.micro.cmd.alert import AlertHandler
    from moat.micro.part.relay import Relay


class BatteryAlert(Alert):
    """
    For a given battery there will at most be one open alert of each type.
    """


@as_proxy("bms_SH")
class HighSOC(BatteryAlert):
    "Charge exceeds limit"


@as_proxy("bms_SL")
class LowSOC(BatteryAlert):
    "Charge below limit"


@as_proxy("bms_UH")
class HighVoltage(BatteryAlert):
    "Total voltage exceeds limit"


@as_proxy("bms_UL")
class LowVoltage(BatteryAlert):
    "Total voltage below limit"


@as_proxy("bms_CH")
class HighCellVoltage(BatteryAlert):
    "Cell voltage exceeds limit"


@as_proxy("bms_CL")
class LowCellVoltage(BatteryAlert):
    "Cell voltage below limit"


@as_proxy("bms_CI")
class CellImbalance(BatteryAlert):
    "Excessive cell imbalance"


@as_proxy("bms_UD")
class VoltageDelta(BatteryAlert):
    "High diff between cell sum and total voltage"


@as_proxy("bms_IH")
class HighChargeCurrent(BatteryAlert):
    "Charge current exceeds limit"


@as_proxy("bms_IL")
class HighDischargeCurrent(BatteryAlert):
    "Discharge current exceeds limit"


@as_proxy("bms_TH")
class HighTemperature(BatteryAlert):
    "Temperature exceeds limit"


@as_proxy("bms_TL")
class LowTemperature(BatteryAlert):
    "Temperature below operational limit"


@as_proxy("bms_NBD")
class NoBatteryData(BatteryAlert):
    "no battery data seen"


@as_proxy("bms_NCD")
class NoCellData(BatteryAlert):
    "no cell data seen"


@as_proxy("bms_EL")
class EnergyLow(BatteryAlert):
    "total energy below zero"


@as_proxy("bms_EH")
class EnergyHigh(BatteryAlert):
    "total energy larger than Wmax"


def _s(r):
    # Helper to create a sequence
    return (x for x in r if x is not None)


class BaseCell(BaseCmd):
    """
    Skeleton for a single cell.

    Configuration::

        lim:
          u:
            abs:
              max: 3.65
              min: 2.55
            ext:
              max: 3.55
              min: 2.9
            std:
              max: 3.5
              min: 2.95
          t:
            abs:
              min: 0
              max: 45
            ext:
              min: 5
              max: 40
            std:
              min: 10
              max: 35
          bal:
            abs: 80
            ext: 70
            std: 60

    The limits are:

    * abs: exceed this and the relay triggers.
    * ext: outside of these the current limit is zero.
    * std: normal operation within this range

    The balancer doesn't have minimum temperatures.
    Comtrollers are free to implement more complicated curves.

    """

    vchg: float | None = None
    vdis: float | None = None
    fchg: bool = False
    fdis: bool = False

    async def cmd_c(self):
        "read cell charge level [0…1]"
        raise NotImplementedError("no idea how")

    async def cmd_u(self):
        "read cell voltage"
        raise NotImplementedError("no idea how")

    async def cmd_t(self):
        "read cell temperature"
        raise NotImplementedError("no idea how")

    async def cmd_tb(self):
        "read cell balancer temperature"
        raise NotImplementedError("no idea how")

    doc_dis = dict(
        _d="bal discharge",
        v="float:threshold",
        f="bool:bypass balancer",
        _r=["float:current thresh", "bool:current bypass"],
    )

    async def cmd_dis(self, v: float | None = None, f: bool | None = None):
        """
        balance low: drain cell until below @v.

        If @f (force) is set, the cell will be skipped by the balancer;
        changing @v will be ignored if @f is `None`.

        Returns (voltage,force) tuple.
        """
        if f is not None:
            self.fdis = f
        if v is not None:
            if not self.fdis or f:
                self.vdis = v
                await self.set_dis()
        return (self.vdis, self.fdis)

    async def set_dis(self):
        """set discharging balancer"""
        raise NotImplementedError("no idea how")

    doc_chg = dict(
        _d="bal charge",
        v="float:threshold",
        f="bool:bypass balancer",
        _r=["float:current thresh", "bool:current bypass"],
    )

    async def cmd_chg(self, v: float | None = None, f: bool | None = None):
        """
        balance high: charge cell until above @v

        Obviously this only works with an active balancer.

        If @force is set, the cell will be skipped by the balancer.

        Returns (voltage,force) tuple.
        """
        if f is not None:
            self.fchg = f
        if v is not None:
            if not self.fchg or f:
                self.vchg = v
                await self.set_chg()
        return (self.vchg, self.fchg)

    async def set_chg(self):
        """set charging balancer"""
        raise NotImplementedError("no idea how")

    doc_lim = dict(
        _d="get limits",
        soc="float:current SoC,unknown=0.5",
        _r=["float:charge pct", "float:discharge pct"],
    )

    async def cmd_lim(self, soc: float = 0.5):
        """
        Limits (as fraction of C) depending on @soc, cell
        temperature, voltage, and whatnot.

        May raise an exception, in which case the relay should disengage.

        Otherwise, returns (chg,dis) factors to limit max cell charge/discharge.
        """
        u = await self.cmd_u()
        t = await self.cmd_t()

        lu = self.cfg["lim"]["u"]
        lt = self.cfg["lim"]["t"]

        if t is None:  # some cells might not have a sensor
            tf = 1
        else:
            if t > lt["abs"]["max"]:
                raise HighTemperature(t, self.path)
            if t < lt["abs"]["min"]:
                raise LowTemperature(t, self.path)

            tf = val2pos(lt["abs"]["max"], t, lt["ext"]["max"], clamp=True) * val2pos(
                lt["abs"]["min"],
                t,
                lt["ext"]["min"],
                clamp=True,
            )

        if u > lu["abs"]["max"]:
            raise HighCellVoltage(u, self.path)
        chg = tf * val2pos(lu["ext"]["max"], u, lu["std"]["max"], clamp=True)

        if u < lu["abs"]["min"]:
            raise LowCellVoltage(u, self.path)
        dis = tf * val2pos(lu["ext"]["min"], u, lu["std"]["min"], clamp=True)

        # TODO use an exponent != 1
        lc = self.cfg["lim"]["c"]
        if soc < lc["min"]:
            dis *= soc / lc["min"]
        if soc > lc["max"]:
            chg *= (1 - soc) / (1 - lc["max"])
        return (chg, dis)


class BalBaseCell(BaseCell):
    "A BaseCell with balancing state"

    in_balance: bool = False
    balance_pwm: float = None  # percentage of time the balancer is on
    balance_over_temp: bool = False
    balance_threshold: float = None
    balance_forced: bool = False

    doc_bal = dict(
        _d="get balancer state",
        _r=dict(
            b="bool:OK",
            f="bool:forced",
            ot="bool:overtemp",
            pwm="dict:pwm data",
            th="float:threshold",
        ),
    )

    async def cmd_bal(self):
        "Get Balancer state/data"
        res = dict(b=self.in_balance, f=self.balance_forced, ot=self.balance_over_temp)
        if self.balance_pwm is not None:
            res["pwm"] = self.balance_pwm
        if self.balance_threshold is not None:
            res["th"] = self.balance_threshold
        return res


class BaseCells(ArrayCmd):
    """
    Skeleton for a cell array.

    Basic configuration::

        apps:
          a: bms._test.Cells
        a:
          app: bms._test.Cell
          cfg: {}
          n: 8
          w: 1000000 # battery's total energy, in watt seconds
          lim:
            ud: 0.02 # relativ difference u/ud
          t:
            w: 500  # calculate cumulative energy every this-many-msec
            ud: 5000  # watch the difference between cell and cumulative voltage
          alarm: !P path.to.alarm.handler
          loss: 0.01 # average capacity loss when charging

    """

    w: attrdict = None
    w_max: float = None
    p: float = None
    al: SubMsgSender[AlertHandler] | None = None
    rly: SubMsgSender[Relay] | None = None
    n_warn_w = 0
    n_warn_ud = 0
    n_save = 98

    def __init__(self, cfg):
        super().__init__(cfg)
        self._reloaded = Event()
        self.clear_work()

    async def _setup(self):
        cfg = self.cfg
        self.al = self.root.sub_at(cfg["alarm"]) if "alarm" in cfg else None
        self.rly = self.root.sub_at(cfg["rly"]) if "relay" in cfg else None
        try:
            self.ud_max = cfg["lim"]["ud"]
        except KeyError:
            self.ud_max = 0.02
        try:
            self.t_w = cfg["t"]["w"]
        except KeyError:
            self.t_w = 500

        c = self.cfg["cfg"]["lim"]["u"]["ext"]
        n = self.cfg["n"]
        self.u_mid = (c["min"] + c["max"]) * n / 2

    async def setup(self):
        await super().setup()
        await self._setup()
        try:
            w = rtc_state[("state",) + self.path]
        except KeyError:
            pass
        else:
            self.work = attrdict(**w)

    doc_c = dict(_d="charge state", _r="float")

    async def cmd_c(self):
        """fetch charge state"""
        if self.w_max is None:
            return 0.5
        return self.w / self.w_max

    doc_u = dict(_d="voltage sum", _r="float")

    async def cmd_u(self):
        """fetch voltage sum"""
        r = await self.cmd_all("u")
        return sum(r)

    doc_t = dict(_d="min/max bat temp", _r=["float:min", "float:max"])

    async def cmd_t(self):
        """fetch temperature min/max"""
        r = await self.cmd_all("t")
        return min(_s(r), default=None), max(_s(r), default=None)

    doc_tb = dict(_d="min/max balancer temp", _r=["float:min", "float:max"])

    async def cmd_tb(self):
        """fetch balancer temperature min/max"""
        r = await self.cmd_all("tb")
        return min(_s(r), default=None), max(_s(r), default=None)

    doc_lim = dict(_d="chg/dischg limit factors", _r=["float:chg", "float:dischg"])

    async def cmd_lim(self):
        """return charge,discharge limit factors"""
        chg, dis = None, None
        try:
            for c, d in await self.cmd_all("lim"):
                if chg is None or chg > c:
                    chg = c
                if dis is None or dis > d:
                    dis = d
            return chg, dis

        except BatteryAlert:
            if self.rly is not None:
                await self.rly.w(v=False)
            raise

    async def reload(self):
        await super().reload()
        self._reloaded.set()
        self._reloaded = Event()

    async def monitor_ud(self):
        while True:
            try:
                await sleep_ms(self.cfg["t"]["ud"])
            except KeyError:
                await self._reloaded.wait()
                continue
            u = await self.cmd_u()
            ud = sum(await self.cmd_all("u"))
            if abs((u - ud) / ud) > self.ud_max:
                if self.al and not (self.n_warn_ud % 10):
                    await self.al.w(a=VoltageDelta, p=self.path, d=dict(u=u, ud=ud))
                self.n_warn_ud += 1
            elif self.n_warn_ud:
                self.n_warn_ud = 0
                await self.al.w(a=VoltageDelta, p=self.path)

    async def get_energy(self):
        """
        loop to calculate battery energy

        The base version sums up (u*i), i.e. watt seconds.
        """
        t1 = ticks_ms()
        i = await self.cmd_i()
        while True:
            await sleep_ms(self.t_w)
            u = await self.cmd_u()
            t2 = ticks_ms()
            self.p = p = i * u
            await self._add_w(p, ticks_diff(t2, t1) / 1000, u > self.u_mid)

            await sleep_ms(self.t_w)
            i = await self.cmd_i()
            t1 = ticks_ms()
            self.p = p = i * u
            await self._add_w(p, ticks_diff(t1, t2) / 1000, u > self.u_mid)

    doc_w = dict(_d="energy content", _r="float:current total", w="float:override")

    async def cmd_w(self, w: float | None = None) -> float:
        """get, or manually override, battery energy content"""
        wx = self.w
        if w is not None:
            self.clear_work(w)
        return wx

    def clear_work(self, w=None):
        if w is None:
            pass
        elif w < 0 or (self.w_max is not None and w > self.w_max):
            return False
        self.work = attrdict()
        self.work.sum = w or 0
        self.work.t = 0
        self.work.chg = 0
        self.work.dis = 0
        self.work.xchg = 0
        self.work.xdis = 0
        return True

    async def _add_w(self, w, t, hi):
        hi  # noqa:B018
        self.work.t += t
        w *= t

        if w > 0:
            self.work.chg += w
            w *= 1 - self.cfg.get("loss", 0)
        else:
            self.work.dis -= w
        self.work.sum += w

        # outside "standard" limits
        if self.work.sum < 0:
            self.work.xdis += -self.work.sum
            self.work.sum = 0
            if w < 0:
                if self.al and not (self.n_warn_w % 10):
                    await self.al.w(a=EnergyLow, p=self.path, d=dict(w=self.work.xdis, wd=w))
                self.n_warn_w += 1

        elif self.w_max is not None and self.work.sum > self.w_max:
            self.work.xchg += self.work.sum - self.w_max
            self.work.sum = self.w_max

            if w > 0:
                if self.al and not (self.n_warn_w % 10):
                    await self.al.w(a=EnergyHigh, p=self.path, d=dict(w=self.work.xchg, wd=w))
                self.n_warn_w += 1

        elif self.n_warn_w:
            self.n_warn_w = 0
            if self.w_max is None or self.work.sum < self.w_max / 2:
                await self.al.w(
                    a=EnergyLow if self.work.sum < self.w_max / 2 else EnergyHigh,
                    p=self.path,
                )

        self.n_save += 1
        if self.n_save > 99:
            self.n_save = 0
            rtc_state[("state",) + self.path] = self.work

    def get_work(self, clear: bool = False, poll: bool = False):
        poll  # noqa:B018
        res = self.work
        if clear:
            self.clear_work()
        return res

    async def start_tasks(self, tg):
        await tg.spawn(self.get_energy)
        await tg.spawn(self.monitor_ud)

    async def task(self):
        async with TaskGroup() as tg:
            await self.start_tasks(tg)
            await super().task()


class BaseBattery(BaseCells):
    """
    Skeleton for a balanced battery.
    """

    # superclass does sum of cell voltages
    #   async def cmd_u(self):
    #       """fetch battery voltage"""
    #       raise RuntimeError

    async def cmd_i(self):
        """fetch battery current"""
        raise RuntimeError

    doc_ud = dict(_d="delta current/total V", _r="float:sum(cell)-total")

    async def cmd_ud(self):
        """Get delta between cell voltage sum and battery voltage."""
        u1 = await self.cmd_u()
        u2 = sum(await self.all("u"))
        return u2 - u1


class BaseBalancer(BaseCmd):
    """
    Basic battery balancing.

    Config::
        bat: !P battery.to.balance
        t:
          chk: 1000  # time between checks when not balancing
          run: 500  # time between updates when balancing
        h:
          n: 3 # max parallel balancing
          d: 0.1  # min absolute voltage delta high vs. low
        u:
          max: 3.45  # no discharging below this point
          min: 3.05  # no charging above this point
    """

    # currently configured limits
    uh: float = None
    ul: float = None

    # configured limits, battery
    chg_max: float = None
    chg_min: float = None
    dis_max: float = None
    dis_min: float = None

    def __init__(self, cfg):
        super().__init__(cfg)
        self._reloaded = Event()
        self._run = Event()

    async def reload(self):
        await super().reload()
        await self._setup()

    async def _setup(self):
        self.n = self.cfg.get("n", 9999)
        self.bat = self.root.sub_at(self.cfg["bat"]) if "bat" in self.cfg else None
        if self.bat is not None:
            # get battery limits
            c = await self.bat.cfg_(("cfg", "lim", "u", "ext"))
            self.dis_max = c["max"]
            self.chg_min = c["min"]
            c = await self.bat.cfg_(("cfg", "lim", "u", "std"))
            self.dis_min = c["max"]
            self.chg_max = c["min"]
            c = self.cfg.get("u", {})
            self.dis_min = max(self.dis_min, c.get("max", 0))
            self.chg_max = max(self.chg_max, c.get("min", 9999))
            if self.dis_min >= self.dis_max:
                raise RuntimeError("Cf dis")
            if self.chg_min >= self.chg_max:
                raise RuntimeError("Cf Chg")

            self._reloaded.set()
            self._reloaded = Event()

    async def setup(self):
        await super().setup()
        await self._setup()

    async def start_tasks(self, tg):
        await tg.spawn(self.run_bms)

    async def task(self):
        async with TaskGroup() as tg:
            await self.start_tasks(tg)
            await super().task()

    doc_u = dict(_d="V limits", h="float:max", l="float:min")

    async def cmd_u(self, h: float | None = None, l: float | None = None):  # noqa:E741
        "set desired voltage levels"
        if h is not None:
            self.uh = min(self.dis_max, max(self.dis_min, h))
        if l is not None:
            self.ul = min(self.chg_max, max(self.chg_min, l))
        self._run.set()

    async def run_bms(self):
        res = False
        while True:
            try:
                await wait_for_ms(self.cfg["t"]["run" if res else "chk"], self._run.wait)
            except KeyError:
                await self._reloaded.wait()
                continue
            except TimeoutError:
                pass
            if self.bat is None:
                await self._reloaded.wait()
                continue

            u = await self.bat.all("u")
            res = await self._run_h(u)
            # res = (await self._run_l(u)) || res

    async def _run_h(self, u):
        maxv = max(u)
        minv = min(u)

        # discharger states
        st = await self.bat.all("dis")

        if not self.uh or self.uh > maxv:
            for uv, f in st:
                if uv and not f:
                    await self.bat.all("dis", v=0)
                    break
            return
        try:
            d = self.cfg["h"]["d"]
        except KeyError:
            d = 0.05

        if maxv - minv < 2 * d or maxv < self.uh:
            # all OK. Don't do any (more) work.
            log("Bal- %.3f %.3f %.3f %.3f", minv, maxv, d, self.dis_min)
            await self.bat.all("dis", v=0)
            return

        thrv1 = self.uh
        minv = max(minv, self.dis_min)
        thrv2 = minv + 0.8 * (thrv1 - minv)  # small hysteresis
        cc = list(enumerate(u))
        cc.sort(key=lambda x: x[1], reverse=True)
        ret = False

        log("Bal %.3f %.3f %s %s", thrv1, thrv2, cc[0], cc[-1])

        want = 0
        cur = 0
        # Step 1, count what we currently have+need (and do some cleanup)
        for i, cv in cc:
            if st[i][1]:
                cur += 1
                continue

            if st[i][0] is not None:
                if st[i][0] < cv:
                    cur += 1
                    if st[i][0] < minv:
                        # don't balance below the minimum
                        log("Balance1 %s", cv)
                        await self.bat(i, "dis", v=cv)
                        # don't spam the system when the min level changes
                else:
                    # goal reached.
                    log("Unbalance1 %s", cv)
                    await self.bat(i, "dis", v=0)

            elif cv >= thrv1 + d:
                want += 1

        #       if cur:
        #           # update balancer power levels
        #           h, res = await self.send(RequestBalancePower())
        #           for c, r in zip(self.cells, res):
        #               r.to_cell(c)

        if cur or want:
            ret = True

        # Step 1, if there are too many active cells, reduce the load
        for i, cv in cc[::-1]:
            if want + cur <= self.n:
                # done
                break
            if st[i][1]:
                continue
            if st[i][0] > cv and cv < thrv2:
                log("Unbalance2 %s", cv)
                await self.bat(i, "dis", v=0)
                cur -= 1

        # Step 2, turn on balancing on cells that need it
        for i, cv in cc:
            if st[i][1]:
                continue
            if st[i][0]:
                continue
            if cur > self.n:
                break
            if cv >= thrv1 + d:
                log("Balance2 %d %s", i, cv)
                await self.bat(i, "dis", v=minv)
                cur += 1
                ret = True
                continue
            break

        return ret



from __future__ import annotations  # noqa: D100

# package("app/bms", opt=0)
# package("moat/ems/battery", opt=0)

module("serialpacker.py", opt=0)
module("async_queue.py", opt=0)
package("moat", opt=0)
package("app", opt=0)
